import json
import logging
import hashlib
from datetime import datetime, UTC
from pathlib import Path
from typing import Dict, List, Optional, Tuple
from dataclasses import dataclass, field

from .security_utils import (
    ScanResult,
    MCPServerInfo,
    ToolInfo,
    ResourceInfo,
    PromptInfo,
    ResourceTemplateInfo,
    safe_load_file,
    MAX_BASELINE_FILE_SIZE
)

logger = logging.getLogger(__name__)

_baseline_cache: Dict[str, 'BaselineData'] = {}


@dataclass(frozen=True)
class BaselineTool:
    name: str
    server_name: str
    config_file: str
    description_hash: str
    metadata_hash: str
    args_hash: str


@dataclass(frozen=True)
class BaselineResource:
    uri: str
    name: str
    server_name: str
    config_file: str
    description_hash: str
    uri_hash: str
    mime_type_hash: str


@dataclass(frozen=True)
class BaselineServer:
    name: str
    config_file: str


@dataclass(frozen=True)
class BaselinePrompt:
    name: str
    server_name: str
    config_file: str
    description_hash: str
    arguments_hash: str
    metadata_hash: str


@dataclass(frozen=True)
class BaselineResourceTemplate:
    uri_template: str
    name: str
    server_name: str
    config_file: str
    description_hash: str
    uri_template_hash: str
    mime_type_hash: str
    metadata_hash: str


@dataclass(frozen=True)
class BaselineData:
    timestamp: str = ""
    config_files: Tuple[str, ...] = field(default_factory=tuple)
    servers: Tuple[BaselineServer, ...] = field(default_factory=tuple)
    tools: Tuple[BaselineTool, ...] = field(default_factory=tuple)
    resources: Tuple[BaselineResource, ...] = field(default_factory=tuple)
    prompts: Tuple[BaselinePrompt, ...] = field(default_factory=tuple)
    resource_templates: Tuple[BaselineResourceTemplate, ...] = field(default_factory=tuple)


def calculate_hash(data: str) -> str:
    return hashlib.sha512(data.encode('utf-8')).hexdigest()


def hash_tool_description(tool: ToolInfo) -> str:
    return calculate_hash(tool.description or "")


def hash_tool_metadata(tool: ToolInfo) -> str:
    metadata = {
        "input_schema": tool.input_schema or {},
        "output_schema": tool.output_schema or {},
        "tags": tool.tags or []
    }
    return calculate_hash(json.dumps(metadata, sort_keys=True))


def hash_tool_args(tool: ToolInfo) -> str:
    input_schema = tool.input_schema or {}
    properties = input_schema.get("properties", {})
    return calculate_hash(json.dumps(properties, sort_keys=True))


def hash_resource_description(resource: ResourceInfo) -> str:
    return calculate_hash(resource.description or "")


def hash_resource_uri(resource: ResourceInfo) -> str:
    return calculate_hash(resource.uri or "")


def hash_resource_mime_type(resource: ResourceInfo) -> str:
    return calculate_hash(resource.mime_type or "")


def hash_prompt_description(prompt: PromptInfo) -> str:
    return calculate_hash(prompt.description or "")


def hash_prompt_arguments(prompt: PromptInfo) -> str:
    return calculate_hash(json.dumps(prompt.arguments or {}, sort_keys=True))


def hash_prompt_metadata(prompt: PromptInfo) -> str:
    metadata = {
        "title": prompt.title or "",
        "tags": prompt.tags or []
    }
    return calculate_hash(json.dumps(metadata, sort_keys=True))


def hash_resource_template_description(template: ResourceTemplateInfo) -> str:
    return calculate_hash(template.description or "")


def hash_resource_template_uri_template(template: ResourceTemplateInfo) -> str:
    return calculate_hash(template.uri_template or "")


def hash_resource_template_mime_type(template: ResourceTemplateInfo) -> str:
    return calculate_hash(template.mime_type or "")


def hash_resource_template_metadata(template: ResourceTemplateInfo) -> str:
    metadata = {
        "title": template.title or "",
        "tags": template.tags or []
    }
    return calculate_hash(json.dumps(metadata, sort_keys=True))


def generate_baseline(scan_result: ScanResult) -> BaselineData:
    servers = []
    for server in scan_result.servers:
        config_file = server.source_file or ""
        servers.append(BaselineServer(
            name=server.name,
            config_file=config_file
        ))
    
    tools = []
    for tool in scan_result.tools:
        config_file = tool.config_file or ""
        tools.append(BaselineTool(
            name=tool.name,
            server_name=tool.server_name,
            config_file=config_file,
            description_hash=hash_tool_description(tool),
            metadata_hash=hash_tool_metadata(tool),
            args_hash=hash_tool_args(tool)
        ))
    
    resources = []
    for resource in scan_result.resources:
        config_file = resource.config_file or ""
        resources.append(BaselineResource(
            uri=resource.uri,
            name=resource.name,
            server_name=resource.server_name,
            config_file=config_file,
            description_hash=hash_resource_description(resource),
            uri_hash=hash_resource_uri(resource),
            mime_type_hash=hash_resource_mime_type(resource)
        ))
    
    prompts = []
    for prompt in scan_result.prompts:
        config_file = prompt.config_file or ""
        prompts.append(BaselinePrompt(
            name=prompt.name,
            server_name=prompt.server_name,
            config_file=config_file,
            description_hash=hash_prompt_description(prompt),
            arguments_hash=hash_prompt_arguments(prompt),
            metadata_hash=hash_prompt_metadata(prompt)
        ))
    
    resource_templates = []
    for template in scan_result.resource_templates:
        config_file = template.config_file or ""
        resource_templates.append(BaselineResourceTemplate(
            uri_template=template.uri_template,
            name=template.name,
            server_name=template.server_name,
            config_file=config_file,
            description_hash=hash_resource_template_description(template),
            uri_template_hash=hash_resource_template_uri_template(template),
            mime_type_hash=hash_resource_template_mime_type(template),
            metadata_hash=hash_resource_template_metadata(template)
        ))
    
    return BaselineData(
        timestamp=datetime.now(UTC).isoformat(),
        config_files=tuple(scan_result.config_files or []),
        servers=tuple(servers),
        tools=tuple(tools),
        resources=tuple(resources),
        prompts=tuple(prompts),
        resource_templates=tuple(resource_templates)
    )


def save_baseline(baseline_data: BaselineData, file_path: Path) -> None:
    baseline_dict = {
        "timestamp": baseline_data.timestamp,
        "config_files": list(baseline_data.config_files),
        "servers": [
            {
                "name": s.name,
                "config_file": s.config_file
            }
            for s in baseline_data.servers
        ],
        "tools": [
            {
                "name": t.name,
                "server_name": t.server_name,
                "config_file": t.config_file,
                "description_hash": t.description_hash,
                "metadata_hash": t.metadata_hash,
                "args_hash": t.args_hash
            }
            for t in baseline_data.tools
        ],
        "resources": [
            {
                "uri": r.uri,
                "name": r.name,
                "server_name": r.server_name,
                "config_file": r.config_file,
                "description_hash": r.description_hash,
                "uri_hash": r.uri_hash,
                "mime_type_hash": r.mime_type_hash
            }
            for r in baseline_data.resources
        ],
        "prompts": [
            {
                "name": p.name,
                "server_name": p.server_name,
                "config_file": p.config_file,
                "description_hash": p.description_hash,
                "arguments_hash": p.arguments_hash,
                "metadata_hash": p.metadata_hash
            }
            for p in baseline_data.prompts
        ],
        "resource_templates": [
            {
                "uri_template": rt.uri_template,
                "name": rt.name,
                "server_name": rt.server_name,
                "config_file": rt.config_file,
                "description_hash": rt.description_hash,
                "uri_template_hash": rt.uri_template_hash,
                "mime_type_hash": rt.mime_type_hash,
                "metadata_hash": rt.metadata_hash
            }
            for rt in baseline_data.resource_templates
        ]
    }
    
    with open(file_path, 'w') as f:
        json.dump(baseline_dict, f, indent=2)
    
    logger.info(f"Baseline saved to {file_path}")


def load_baseline(file_path: Path) -> Optional[BaselineData]:
    global _baseline_cache
    
    abs_path = str(file_path.resolve())
    
    if abs_path in _baseline_cache:
        logger.debug(f"Using cached baseline from {abs_path}")
        return _baseline_cache[abs_path]
    
    if not file_path.exists():
        return None
    
    baseline_data = _load_baseline_from_file(file_path)
    if baseline_data is None:
        return None
    
    _baseline_cache[abs_path] = baseline_data
    logger.debug(f"Cached baseline from {abs_path}")
    
    return baseline_data


def _load_baseline_from_file(file_path: Path) -> Optional[BaselineData]:
    try:
        data = safe_load_file(file_path, MAX_BASELINE_FILE_SIZE)
        
        servers = [
            BaselineServer(
                name=s["name"],
                config_file=s.get("config_file", "")
            )
            for s in data.get("servers", [])
        ]
        
        tools = [
            BaselineTool(
                name=t["name"],
                server_name=t["server_name"],
                config_file=t.get("config_file", ""),
                description_hash=t["description_hash"],
                metadata_hash=t["metadata_hash"],
                args_hash=t["args_hash"]
            )
            for t in data.get("tools", [])
        ]
        
        resources = [
            BaselineResource(
                uri=r["uri"],
                name=r["name"],
                server_name=r["server_name"],
                config_file=r.get("config_file", ""),
                description_hash=r["description_hash"],
                uri_hash=r["uri_hash"],
                mime_type_hash=r["mime_type_hash"]
            )
            for r in data.get("resources", [])
        ]
        
        prompts = [
            BaselinePrompt(
                name=p["name"],
                server_name=p["server_name"],
                config_file=p.get("config_file", ""),
                description_hash=p["description_hash"],
                arguments_hash=p["arguments_hash"],
                metadata_hash=p["metadata_hash"]
            )
            for p in data.get("prompts", [])
        ]
        
        resource_templates = [
            BaselineResourceTemplate(
                uri_template=rt["uri_template"],
                name=rt["name"],
                server_name=rt["server_name"],
                config_file=rt.get("config_file", ""),
                description_hash=rt["description_hash"],
                uri_template_hash=rt["uri_template_hash"],
                mime_type_hash=rt["mime_type_hash"],
                metadata_hash=rt["metadata_hash"]
            )
            for rt in data.get("resource_templates", [])
        ]
        
        return BaselineData(
            timestamp=data.get("timestamp", ""),
            config_files=tuple(data.get("config_files", [])),
            servers=tuple(servers),
            tools=tuple(tools),
            resources=tuple(resources),
            prompts=tuple(prompts),
            resource_templates=tuple(resource_templates)
        )
    except ValueError as e:
        logger.error(f"Baseline file size limit exceeded: {e}")
        return None
    except (KeyError, json.JSONDecodeError, TypeError) as e:
        logger.error(f"Invalid baseline file format: {e}")
        return None
    except Exception as e:
        logger.error(f"Failed to load baseline from {file_path}: {e}")
        return None




import asyncio
import json
import logging
import re
from typing import List, Optional, Set, Pattern, Tuple

from ..security_utils import (
    ScanResult,
    SecurityIssue,
    Severity,
    create_security_issue,
    extract_tool_content,
    extract_resource_content,
    extract_prompt_content,
    extract_resource_template_content,
    extract_text_content,
)

logger = logging.getLogger(__name__)

critical_patterns: List[str] = [
    r"\$\(.*?\)",
    r"`.*?`",
    r"\$\{.*?\}",
    r"\b(exec|system|shell_exec|passthru|popen|proc_open)\b",
    r"\b(fork|vfork|clone)\b",
    r"[;&|]\s*\w",
    r"\b(&&|\|\|)\b",
]

context_patterns: List[str] = [
    r"\b(sh|bash|zsh|ksh|csh|tcsh|fish|powershell|pwsh|cmd\.exe|command\.com)\s+(?:-c|-e|--execute)",
    r"\b(python[0-9.]*|python3|py|perl|ruby|node|nodejs|php|java|go|rust)\s+(?:-c|-e|--eval|--execute|run)",
    r"\b(cat|curl|wget|nc|netcat|ncat|telnet|ftp|sftp|scp|ssh|rsync)\b.*[;&|>]",
    r"\b(chmod|chown|rm|mv|cp|mkdir|rmdir|touch|ln)\b.*[;&|>]",
    r"\b(sudo|su)\s+\w+",
    r"\b(kill|killall|pkill)\s+\w+",
    r"\b(mount|umount|fdisk|mkfs|dd)\s+[^\s]+",
    r"\b(systemctl|service|initctl)\s+\w+",
    r"\b(crontab|at|batch)\s+[^\s]+",
    r"\b(tee|sponge)\b.*[>|&]",
    r"\b(ping|traceroute|tracert|nslookup|dig|whois)\b.*[;&|]",
    r"\b(nmap|masscan|zmap|unicornscan)\s+[^\s]+",
    r"\b(powershell|pwsh|wsl|bash|sh)\s+(?:-c|-e)",
    r"\b(mysql|psql|sqlite|mongo|redis-cli|memcached)\b.*[;&|]",
    r"\b(docker|podman|kubectl)\s+(?:exec|run|exec)",
    r"\b(reg|regedit|regsvr32|rundll32|wscript|cscript)\s+[^\s]+",
    r"\b(taskkill|schtasks|wmic)\s+[^\s]+",
]

benign_patterns: List[str] = [
    r"(?:written|developed|coded)\s+in\s+(python|java|go|rust|perl|ruby|node)",
    r"(?:uses|utilizes)\s+(curl|wget|http)\s+(?:to|for)",
    r"(?:example|demo|sample)\s+(?:command|usage|syntax)",
    r"(?:see|refer|check)\s+(?:documentation|docs|readme|manual)",
    r"https?://[\w\-\.]+",
    r"[\w\-\.]+@[\w\-\.]+\.\w+",
]

safe_string = re.compile(r"^[\w\s\-._@:/?#\[\]]+$")

_ENTITY_BATCH_SIZE = 16

_COMPILED_CRITICAL_PATTERNS: Optional[List[Pattern]] = None
_COMPILED_CONTEXT_PATTERNS: Optional[List[Pattern]] = None
_COMPILED_BENIGN_PATTERNS: Optional[List[Pattern]] = None


def _get_compiled_patterns() -> Tuple[List[Pattern], List[Pattern], List[Pattern]]:
    global _COMPILED_CRITICAL_PATTERNS, _COMPILED_CONTEXT_PATTERNS, _COMPILED_BENIGN_PATTERNS
    
    if _COMPILED_CRITICAL_PATTERNS is None:
        _COMPILED_CRITICAL_PATTERNS = []
        for pattern_str in critical_patterns:
            try:
                _COMPILED_CRITICAL_PATTERNS.append(re.compile(pattern_str, re.IGNORECASE))
            except re.error as e:
                logger.warning(f"Failed to compile critical pattern '{pattern_str}': {e}")
        
        _COMPILED_CONTEXT_PATTERNS = []
        for pattern_str in context_patterns:
            try:
                _COMPILED_CONTEXT_PATTERNS.append(re.compile(pattern_str, re.IGNORECASE))
            except re.error as e:
                logger.warning(f"Failed to compile context pattern '{pattern_str}': {e}")
        
        _COMPILED_BENIGN_PATTERNS = []
        for pattern_str in benign_patterns:
            try:
                _COMPILED_BENIGN_PATTERNS.append(re.compile(pattern_str, re.IGNORECASE))
            except re.error as e:
                logger.warning(f"Failed to compile benign pattern '{pattern_str}': {e}")
    
    return _COMPILED_CRITICAL_PATTERNS, _COMPILED_CONTEXT_PATTERNS, _COMPILED_BENIGN_PATTERNS


def _scan_text_for_commands(text: str) -> Tuple[Set[str], List[str]]:
    if not text:
        return set(), []

    compiled_critical, compiled_context, compiled_benign = _get_compiled_patterns()

    segments = re.split(r"[\n\r]+|\s{2,}", text)
    critical_hits: Set[str] = set()
    context_hits: Set[str] = set()
    benign_hits: Set[str] = set()
    matched_snippets: List[str] = []

    for seg in segments:
        if not isinstance(seg, str):
            continue
        s = seg.strip()
        if not s or safe_string.match(s):
            continue

        for i, compiled_pat in enumerate(compiled_critical):
            match = compiled_pat.search(s)
            if match:
                critical_hits.add(critical_patterns[i])
                start = max(0, match.start() - 50)
                end = min(len(s), match.end() + 50)
                context_snippet = s[start:end].strip()
                if context_snippet and context_snippet not in matched_snippets:
                    matched_snippets.append(context_snippet[:150])

        for i, compiled_pat in enumerate(compiled_context):
            match = compiled_pat.search(s)
            if match:
                context_hits.add(context_patterns[i])
                start = max(0, match.start() - 50)
                end = min(len(s), match.end() + 50)
                context_snippet = s[start:end].strip()
                if context_snippet and context_snippet not in matched_snippets:
                    matched_snippets.append(context_snippet[:150])

    text_lower = text.lower()
    for i, compiled_pat in enumerate(compiled_benign):
        if compiled_pat.search(text_lower):
            benign_hits.add(benign_patterns[i])

    critical_count = len(critical_hits)
    context_count = len(context_hits)
    malicious_count = critical_count + context_count
    benign_count = len(benign_hits)

    all_hits = critical_hits | context_hits

    if critical_count >= 2:
        return all_hits, matched_snippets[:5]

    if critical_count >= 1 and context_count >= 1:
        return all_hits, matched_snippets[:5]

    if malicious_count >= 2 and benign_count < malicious_count:
        return all_hits, matched_snippets[:5]

    if malicious_count >= 3:
        return all_hits, matched_snippets[:5]

    return set(), []


async def _scan_single_tool(tool, servers_list) -> Optional[SecurityIssue]:
    try:
        server = next((s for s in servers_list if s.name == tool.server_name), None)
        config_file = server.source_file if server else None

        tool_content = extract_tool_content(tool)
        text_content = extract_text_content(tool_content)
        
        matches, matched_content = await asyncio.to_thread(_scan_text_for_commands, text_content)

        if matches:
            issue = create_security_issue(
                issue_type="Command Injection",
                severity=Severity.CRITICAL,
                entity_type="tool",
                affected_tool=tool.name,
                affected_server=tool.server_name,
                config_file=config_file,
                affected_entities={
                    "pattern_count": len(matches),
                    "matched_content": matched_content
                }
            )
            logger.warning(
                f"Command injection detected in tool: {tool.name} (found {len(matches)} pattern groups)")
            return issue
        return None
    except Exception as e:
        logger.error(f"Error scanning tool {tool.name} for command injection: {e}")
        return None


async def _scan_single_resource(resource, servers_list) -> Optional[SecurityIssue]:
    try:
        server = next((s for s in servers_list if s.name == resource.server_name), None)
        config_file = server.source_file if server else None

        resource_content = extract_resource_content(resource)
        text_content = extract_text_content(resource_content)
        
        matches, matched_content = await asyncio.to_thread(_scan_text_for_commands, text_content)

        if matches:
            issue = create_security_issue(
                issue_type="Command Injection",
                severity=Severity.CRITICAL,
                entity_type="resource",
                affected_resource=resource.name,
                affected_resource_uri=resource.uri,
                affected_server=resource.server_name,
                config_file=config_file,
                affected_entities={
                    "pattern_count": len(matches),
                    "matched_content": matched_content
                }
            )
            logger.warning(
                f"Command injection detected in resource: {resource.uri} (found {len(matches)} pattern groups)")
            return issue
        return None
    except Exception as e:
        logger.error(f"Error scanning resource {resource.uri} for command injection: {e}")
        return None


async def scan_for_tool_command_injection(scan_result: ScanResult) -> List[SecurityIssue]:
    issues: List[SecurityIssue] = []

    if not scan_result.tools:
        return issues

    logger.info(f"Scanning {len(scan_result.tools)} tools for command injection")

    for i in range(0, len(scan_result.tools), _ENTITY_BATCH_SIZE):
        batch = scan_result.tools[i:i + _ENTITY_BATCH_SIZE]
        tasks = [_scan_single_tool(tool, scan_result.servers) for tool in batch]
        
        batch_results = await asyncio.gather(*tasks, return_exceptions=True)
        
        for result in batch_results:
            if result is not None and not isinstance(result, Exception):
                issues.append(result)
            elif isinstance(result, Exception):
                logger.error(f"Unexpected exception in tool scan: {result}")

    logger.info(f"Tool command injection scan completed: {len(issues)} issues found")
    return issues


async def scan_for_resource_command_injection(scan_result: ScanResult) -> List[SecurityIssue]:
    issues: List[SecurityIssue] = []

    if not scan_result.resources:
        return issues

    logger.info(f"Scanning {len(scan_result.resources)} resources for command injection")

    for i in range(0, len(scan_result.resources), _ENTITY_BATCH_SIZE):
        batch = scan_result.resources[i:i + _ENTITY_BATCH_SIZE]
        tasks = [_scan_single_resource(resource, scan_result.servers) for resource in batch]
        
        batch_results = await asyncio.gather(*tasks, return_exceptions=True)
        
        for result in batch_results:
            if result is not None and not isinstance(result, Exception):
                issues.append(result)
            elif isinstance(result, Exception):
                logger.error(f"Unexpected exception in resource scan: {result}")

    logger.info(f"Resource command injection scan completed: {len(issues)} issues found")
    return issues


async def _scan_single_prompt(prompt, servers_list) -> Optional[SecurityIssue]:
    try:
        server = next((s for s in servers_list if s.name == prompt.server_name), None)
        config_file = server.source_file if server else None

        prompt_content = extract_prompt_content(prompt)
        text_content = extract_text_content(prompt_content)
        
        matches, matched_content = await asyncio.to_thread(_scan_text_for_commands, text_content)

        if matches:
            issue = create_security_issue(
                issue_type="Command Injection",
                severity=Severity.CRITICAL,
                entity_type="prompt",
                affected_server=prompt.server_name,
                config_file=config_file,
                affected_entities={
                    "pattern_count": len(matches),
                    "matched_content": matched_content
                }
            )
            logger.warning(
                f"Command injection detected in prompt: {prompt.name} (found {len(matches)} pattern groups)")
            return issue
        return None
    except Exception as e:
        logger.error(f"Error scanning prompt {prompt.name} for command injection: {e}")
        return None


async def _scan_single_resource_template(template, servers_list) -> Optional[SecurityIssue]:
    try:
        server = next((s for s in servers_list if s.name == template.server_name), None)
        config_file = server.source_file if server else None

        template_content = extract_resource_template_content(template)
        text_content = extract_text_content(template_content)
        
        matches, matched_content = await asyncio.to_thread(_scan_text_for_commands, text_content)

        if matches:
            issue = create_security_issue(
                issue_type="Command Injection",
                severity=Severity.CRITICAL,
                entity_type="resource_template",
                affected_server=template.server_name,
                config_file=config_file,
                affected_entities={
                    "resource_template": template.name,
                    "uri_template": template.uri_template,
                    "pattern_count": len(matches),
                    "matched_content": matched_content
                }
            )
            logger.warning(
                f"Command injection detected in resource template: {template.uri_template} (found {len(matches)} pattern groups)")
            return issue
        return None
    except Exception as e:
        logger.error(f"Error scanning resource template {template.uri_template} for command injection: {e}")
        return None


async def scan_for_prompt_command_injection(scan_result: ScanResult) -> List[SecurityIssue]:
    issues: List[SecurityIssue] = []

    if not scan_result.prompts:
        return issues

    logger.info(f"Scanning {len(scan_result.prompts)} prompts for command injection")

    for i in range(0, len(scan_result.prompts), _ENTITY_BATCH_SIZE):
        batch = scan_result.prompts[i:i + _ENTITY_BATCH_SIZE]
        tasks = [_scan_single_prompt(prompt, scan_result.servers) for prompt in batch]
        
        batch_results = await asyncio.gather(*tasks, return_exceptions=True)
        
        for result in batch_results:
            if result is not None and not isinstance(result, Exception):
                issues.append(result)
            elif isinstance(result, Exception):
                logger.error(f"Unexpected exception in prompt scan: {result}")

    logger.info(f"Prompt command injection scan completed: {len(issues)} issues found")
    return issues


async def scan_for_resource_template_command_injection(scan_result: ScanResult) -> List[SecurityIssue]:
    issues: List[SecurityIssue] = []

    if not scan_result.resource_templates:
        return issues

    logger.info(f"Scanning {len(scan_result.resource_templates)} resource templates for command injection")

    for i in range(0, len(scan_result.resource_templates), _ENTITY_BATCH_SIZE):
        batch = scan_result.resource_templates[i:i + _ENTITY_BATCH_SIZE]
        tasks = [_scan_single_resource_template(template, scan_result.servers) for template in batch]
        
        batch_results = await asyncio.gather(*tasks, return_exceptions=True)
        
        for result in batch_results:
            if result is not None and not isinstance(result, Exception):
                issues.append(result)
            elif isinstance(result, Exception):
                logger.error(f"Unexpected exception in resource template scan: {result}")

    logger.info(f"Resource template command injection scan completed: {len(issues)} issues found")
    return issues


import asyncio
import argparse
import json
import sys
import logging
import os
import shutil
import textwrap
from pathlib import Path
from datetime import datetime
from typing import Optional, List

from rich.console import Console
from rich.text import Text


class SuppressStderr:

    def __enter__(self):
        self.original_stderr = sys.stderr
        sys.stderr = open(os.devnull, 'w')
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stderr.close()
        sys.stderr = self.original_stderr


SEVERITY_EMOJI = {
    'CRITICAL': '🚫',
    'HIGH': '🔴', 
    'MEDIUM': '🟠',
    'LOW': '🟡',
}

RISK_EMOJI = {**SEVERITY_EMOJI, 'NONE': '✅'}


def get_level_emoji(level, mapping):
    level_upper = level.upper() if isinstance(level, str) else level
    return mapping.get(level_upper, '❓')


def print_separator(char='═', length=50):
    print(char * length)


def truncate_lines(text: str, max_lines: int = 2) -> str:
    if not text:
        return ""
    lines = str(text).splitlines()
    if len(lines) <= max_lines:
        return "\n".join(lines)
    return "\n".join(lines[:max_lines]) + "…"


def wrap_text_with_indent(text: str, indent: str = "       ", max_lines: int = None) -> str:
    if not text:
        return ""
    
    try:
        terminal_width = shutil.get_terminal_size().columns
    except (OSError, AttributeError):
        terminal_width = 80
    
    indent_len = len(indent)
    available_width = terminal_width - indent_len
    
    if available_width < 20:
        available_width = 20
    
    wrapped_lines = textwrap.wrap(
        text, 
        width=available_width, 
        break_long_words=False, 
        break_on_hyphens=False
    )
    
    if max_lines and len(wrapped_lines) > max_lines:
        wrapped_lines = wrapped_lines[:max_lines]
        if len(wrapped_lines) == max_lines:
            last_line = wrapped_lines[-1]
            if len(last_line) + 3 <= available_width:
                wrapped_lines[-1] = last_line + "…"
            else:
                max_chars = available_width - 3
                wrapped_lines[-1] = last_line[:max_chars] + "…"
    
    return "\n".join([indent + line for line in wrapped_lines])


def print_banner():
    console = Console()

    banner = """
███    ███  ██████ ██████      █████  ██████  ███    ███  ██████  ██████  
████  ████ ██      ██   ██    ██   ██ ██   ██ ████  ████ ██    ██ ██   ██ 
██ ████ ██ ██      ██████     ███████ ██████  ██ ████ ██ ██    ██ ██████  
██  ██  ██ ██      ██         ██   ██ ██  ██  ██  ██  ██ ██    ██ ██  ██  
██      ██  ██████ ██         ██   ██ ██   ██ ██      ██  ██████  ██   ██ 
"""

    console.print(banner, style="turquoise2")

    banner_text = Text("                                          ｂｙ  Ａｉｒａ  Ｓｅｃｕｒｉｔｙ",
                       style="bold magenta")
    console.print(banner_text)
    console.print()


def print_boxed_section(title, content_func):
    print_separator('═', 60)
    print(f"{title}")
    print_separator('─', 60)
    content_func()
    print()


def _format_entity_display(entity) -> str:
    entity_type_map = {
        "tool": "Tool",
        "resource": "Resource",
        "prompt": "Prompt",
        "resource_template": "Resource Template",
        "server": "Server",
        "configuration": "Configuration"
    }
    entity_label = entity_type_map.get(entity.entity_type, entity.entity_type.title())
    
    changed_fields = entity.changed_fields or []
    changed_str = f" [changed: {', '.join(changed_fields)}]" if changed_fields else ""
    
    disabled = entity.additional_info.get("disabled", False) if entity.additional_info else False
    disabled_str = " [disabled]" if disabled else ""
    
    # Special handling for tool name ambiguity - show both conflicting tools
    if entity.entity_type == "tool" and entity.additional_info and "tools" in entity.additional_info:
        tools_list = entity.additional_info["tools"]
        if isinstance(tools_list, list) and len(tools_list) >= 2:
            tool1 = tools_list[0]
            tool2 = tools_list[1]
            tool1_name = tool1.get("name", "unknown") if isinstance(tool1, dict) else str(tool1)
            tool1_server = tool1.get("server", "unknown") if isinstance(tool1, dict) else "unknown"
            tool2_name = tool2.get("name", "unknown") if isinstance(tool2, dict) else str(tool2)
            tool2_server = tool2.get("server", "unknown") if isinstance(tool2, dict) else "unknown"
            score = entity.additional_info.get("score", "")
            score_str = f" (similarity: {score}%)" if score else ""
            return f"Conflict between '{tool1_name}' (server: {tool1_server}) and '{tool2_name}' (server: {tool2_server}) -{score_str}"
    
    if entity.entity_type == "configuration":
        if entity.config_file:
            return f"Server '{entity.entity_name}' (config: '{entity.config_file}'){disabled_str}{changed_str}"
        else:
            return f"Server '{entity.entity_name}'{disabled_str}{changed_str}"
    
    if entity.entity_type == "resource" and entity.affected_resource_uri:
        return f"{entity_label} '{entity.affected_resource_uri}' (server: '{entity.server_name}'){changed_str}"
    elif entity.entity_type == "resource_template" and entity.affected_resource_template_uri:
        return f"{entity_label} '{entity.affected_resource_template_uri}' (server: '{entity.server_name}'){changed_str}"
    else:
        return f"{entity_label} '{entity.entity_name}' (server: '{entity.server_name}'){changed_str}"


def _entity_to_dict(entity) -> dict:
    result = {
        "entity_type": entity.entity_type,
        "entity_name": entity.entity_name,
        "server_name": entity.server_name,
    }
    
    if entity.config_file:
        result["config_file"] = entity.config_file
    if entity.affected_tool:
        result["affected_tool"] = entity.affected_tool
    if entity.affected_resource:
        result["affected_resource"] = entity.affected_resource
    if entity.affected_resource_uri:
        result["affected_resource_uri"] = entity.affected_resource_uri
    if entity.affected_prompt:
        result["affected_prompt"] = entity.affected_prompt
    if entity.affected_resource_template:
        result["affected_resource_template"] = entity.affected_resource_template
    if entity.affected_resource_template_uri:
        result["affected_resource_template_uri"] = entity.affected_resource_template_uri
    if entity.changed_fields:
        result["changed_fields"] = entity.changed_fields
    if entity.additional_info:
        result["additional_info"] = entity.additional_info
    
    return result


def _grouped_issue_to_dict(grouped) -> dict:

    return {
        "finding_type": grouped.issue_type,
        "severity": grouped.severity.value,
        "description": grouped.description,
        "remediation": grouped.remediation,
        "total_entities": grouped.total_entities,
        "entities_affected": [_entity_to_dict(entity) for entity in grouped.entities_affected]
    }


def _redact_headers(headers: dict) -> dict:
    if not headers:
        return {}
    sensitive_headers = {
        'authorization', 'auth', 'authenticate',
        'api-key', 'api_key', 'apikey', 'apikey', 'apiKey', 'API_KEY', 'API-KEY',
        'token', 'access-token', 'access_token', 'accessToken', 'ACCESS_TOKEN',
        'x-api-key', 'x-api-key', 'X-API-Key', 'X-API-KEY',
        'bearer', 'oauth', 'oauth-token', 'oauth_token',
        'session', 'session-id', 'session_id', 'sessionId',
        'cookie', 'set-cookie',
        'x-auth-token', 'x-auth-token', 'X-Auth-Token',
        'x-access-token', 'x-access-token', 'X-Access-Token'
    }
    return {k: "***" if k.lower() in {h.lower() for h in sensitive_headers} else v for k, v in headers.items()}


def _redact_env(env: dict) -> dict:
    if not env:
        return {}
    sensitive_keywords = {
        'key', 'token', 'secret', 'password', 'passwd', 'pwd',
        'api_key', 'apikey', 'api-key', 'apiKey', 'API_KEY',
        'auth', 'authorization', 'authenticate',
        'credential', 'credentials', 'creds',
        'access', 'access_token', 'access-token', 'accessToken',
        'session', 'session_id', 'session-id', 'sessionId',
        'bearer', 'oauth', 'oauth_token', 'oauth-token',
        'private', 'private_key', 'private-key', 'privateKey',
        'cert', 'certificate', 'cert_key', 'cert-key',
        'ssh', 'ssh_key', 'ssh-key', 'sshKey',
        'aws', 'aws_key', 'aws_key_id', 'aws_secret',
        'gcp', 'gcp_key', 'gcp_credential',
        'azure', 'azure_key', 'azure_secret'
    }
    return {k: "***" if any(keyword in k.lower() for keyword in sensitive_keywords) else v for k, v in env.items()}


def _redact_args(args: List[str]) -> List[str]:
    if not args:
        return []
    
    sensitive_flags = {
        '--api-key', '--apikey', '--api_key', '--apiKey', '--API_KEY', '--API-KEY',
        '--token', '--auth-token', '--access-token', '--access_token', '--accessToken',
        '--bearer', '--oauth', '--oauth-token', '--oauth_token',
        '--password', '--passwd', '--pwd', '--pass',
        '--secret', '--secret-key', '--secretkey', '--secret_key', '--secretKey',
        '--credential', '--credentials', '--creds',
        '--auth', '--authorization', '--authenticate',
        '-k', '--key', '-t', '--token', '-p', '--password', '-s', '--secret',
        '--aws-key', '--aws-secret', '--gcp-key', '--azure-key'
    }
    
    redacted = []
    i = 0
    while i < len(args):
        arg = args[i]
        arg_lower = arg.lower()
        
        if '=' in arg:
            key, value = arg.split('=', 1)
            key_lower = key.lower()
            if any(flag in key_lower for flag in sensitive_flags) or \
               any(sensitive in key_lower for sensitive in ['key', 'token', 'secret', 'password', 'auth', 'credential', 'passwd', 'pwd', 'cred']):
                redacted.append(f"{key}=***")
            else:
                redacted.append(arg)
        elif arg_lower in sensitive_flags:
            redacted.append(arg)
            if i + 1 < len(args) and not args[i + 1].startswith('-'):
                redacted.append("***")
                i += 1
        else:
            redacted.append(arg)
        i += 1
    
    return redacted


def setup_logging(show_logs: bool = False):
    log_dir = Path("logs")
    log_dir.mkdir(exist_ok=True)

    log_file = log_dir / "mcp_armor.log"

    formatter = logging.Formatter(
        '%(asctime)s - %(name)s - %(levelname)s - %(message)s',
        datefmt='%Y-%m-%d %H:%M:%S'
    )

    root_logger = logging.getLogger()
    root_logger.setLevel(logging.INFO)
    root_logger.handlers.clear()

    file_handler = logging.FileHandler(log_file, encoding='utf-8')
    file_handler.setLevel(logging.INFO)
    file_handler.setFormatter(formatter)
    root_logger.addHandler(file_handler)

    if show_logs:
        console_handler = logging.StreamHandler()
        console_handler.setLevel(logging.INFO)
        console_handler.setFormatter(formatter)
        root_logger.addHandler(console_handler)

    return str(log_file)


def format_scan_results_as_markdown(results, baseline_status: Optional[dict] = None) -> str:
    from .security_utils import ScanResult, AffectedEntity, get_generic_description, get_generic_recommendation
    md = []

    md.append("# MCP Armor Scan Report")
    md.append(f"Generated: {datetime.now().strftime('%Y-%m-%d %H:%M:%S')}")
    md.append("")

    md.append("## Summary")
    md.append(f"- **Servers found**: {len(results.servers)}")
    md.append(f"- **Tools found**: {len(results.tools)}")
    md.append(f"- **Resources found**: {len(results.resources)}")
    md.append(f"- **Prompts found**: {len(results.prompts)}")
    md.append(f"- **Resource Templates found**: {len(results.resource_templates)}")
    md.append(f"- **Config files**: {len(results.config_files)}")
    md.append(f"- **Errors**: {len(results.errors)}")
    if results.status:
        md.append(f"- **Status**:")
        for status_msg in results.status:
            md.append(f"  - {status_msg}")
    md.append("")

    if results.servers:
        md.append("## Discovered Servers")
        for server in results.servers:
            status = "DISABLED" if server.disabled else "ENABLED"
            md.append(f"### {server.name} ({server.type.value}) - {status}")
            if server.source_file:
                md.append(f"- **Source**: `{server.source_file}`")
            if server.endpoint:
                md.append(f"- **Endpoint**: `{server.endpoint}`")
            if server.command:
                md.append(f"- **Command**: `{server.command}`")
            if server.args:
                md.append(f"- **Args**: `{' '.join(_redact_args(server.args))}`")
            if server.headers:
                redacted_headers = _redact_headers(server.headers)
                headers_str = ', '.join([f"`{k}={v}`" for k, v in redacted_headers.items()])
                md.append(f"- **Headers**: {headers_str}")
            if server.env:
                redacted_env = _redact_env(server.env)
                env_str = ', '.join([f"`{k}={v}`" for k, v in redacted_env.items()])
                md.append(f"- **Environment Variables**: {env_str}")
            if server.additional_permissions:
                md.append(f"- **Permissions**: {', '.join(server.additional_permissions)}")
            md.append("")

    if results.tools:
        md.append("## Discovered Tools")
        for tool in results.tools:
            md.append(f"### {tool.name}")
            md.append(f"- **Server**: {tool.server_name}")
            md.append(f"- **Description**: {tool.description}")
            if tool.input_schema:
                md.append(f"- **Schema**:")
                md.append("```json")
                md.append(json.dumps(tool.input_schema, indent=2))
                md.append("```")
            md.append("")

    if results.resources:
        md.append("## Discovered Resources")
        for resource in results.resources:
            md.append(f"### {resource.name}")
            md.append(f"- **URI**: `{resource.uri}`")
            md.append(f"- **Server**: {resource.server_name}")
            md.append(f"- **Description**: {resource.description}")
            md.append(f"- **MIME Type**: {resource.mime_type}")
            md.append("")

    if results.prompts:
        md.append("## Discovered Prompts")
        for prompt in results.prompts:
            md.append(f"### {prompt.name}")
            md.append(f"- **Server**: {prompt.server_name}")
            md.append(f"- **Description**: {prompt.description}")
            if prompt.arguments:
                md.append(f"- **Arguments**:")
                md.append("```json")
                md.append(json.dumps(prompt.arguments, indent=2))
                md.append("```")
            md.append("")

    if results.resource_templates:
        md.append("## Discovered Resource Templates")
        for template in results.resource_templates:
            md.append(f"### {template.name}")
            md.append(f"- **URI Template**: `{template.uri_template}`")
            md.append(f"- **Server**: {template.server_name}")
            md.append(f"- **Description**: {template.description}")
            md.append(f"- **MIME Type**: {template.mime_type}")
            md.append("")

    if baseline_status:
        md.append("## Baseline Status")
        md.append(f"- **Baseline File**: `{baseline_status.get('baseline_file', 'N/A')}`")
        md.append(f"- **Baseline Loaded**: {baseline_status.get('baseline_loaded', False)}")
        if 'warning' in baseline_status:
            md.append(f"- **Warning**: {baseline_status['warning']}")
        md.append("")

    if hasattr(results, 'security_report') and results.security_report:
        security_report = results.security_report
        md.append("## Security Report")
        md.append(f"- **Risk Level**: {security_report.risk_level.value}")
        md.append(f"- **Total Findings**: {security_report.total_issues}")
        md.append("")

        md.append("### Findings by Severity")
        for severity, count in security_report.issues_by_severity.items():
            if count > 0:
                md.append(f"- **{severity.value}**: {count}")
        md.append("")

        if security_report.issues_by_type:
            md.append("### Findings by Type")
            for issue_type, count in security_report.issues_by_type.items():
                md.append(f"- **{issue_type}**: {count}")
            md.append("")

        if security_report.grouped_issues:
            md.append("### Security Findings")
            for grouped in security_report.grouped_issues:
                sev_str = grouped.severity.value.upper()
                md.append(f"#### {grouped.issue_type} ({sev_str})")
                md.append(f"**Description**: {grouped.description}")
                md.append("")
                md.append(f"**Entities Affected ({grouped.total_entities})**:")
                for entity in grouped.entities_affected:
                    entity_display = _format_entity_display(entity)
                    md.append(f"- {entity_display}")
                md.append("")
                md.append(f"**Remediation**: {grouped.remediation}")
                md.append("")
                md.append("---")
                md.append("")

    if results.errors:
        md.append("## Errors")
        for error in results.errors:
            md.append(f"- {error}")
        md.append("")

    if results.status:
        md.append("## Status")
        for status_msg in results.status:
            md.append(f"- {status_msg}")
        md.append("")

    return "\n".join(md)


def format_inspect_results_as_markdown(results, baseline_info: Optional[dict] = None) -> str:
    from .security_utils import ScanResult
    md = []

    md.append("# MCP Configuration Inspection Report")
    md.append(f"Generated: {datetime.now().strftime('%Y-%m-%d %H:%M:%S')}")
    md.append("")

    md.append("## Summary")
    md.append(f"- **Servers found**: {len(results.servers)}")
    md.append(f"- **Tools found**: {len(results.tools)}")
    md.append(f"- **Resources found**: {len(results.resources)}")
    md.append(f"- **Prompts found**: {len(results.prompts)}")
    md.append(f"- **Resource Templates found**: {len(results.resource_templates)}")
    md.append(f"- **Config files**: {len(results.config_files)}")
    md.append(f"- **Errors**: {len(results.errors)}")
    if results.status:
        md.append(f"- **Status**:")
        for status_msg in results.status:
            md.append(f"  - {status_msg}")
    md.append("")

    if baseline_info:
        md.append("## Baseline")
        md.append(f"- **Baseline File**: `{baseline_info.get('baseline_file', 'N/A')}`")
        md.append(f"- **Baseline Generated**: {baseline_info.get('baseline_generated', False)}")
        if baseline_info.get('timestamp'):
            md.append(f"- **Generated At**: {baseline_info['timestamp']}")
        md.append("")

    if results.config_files:
        md.append("## Configuration Files")
        for config_file in results.config_files:
            md.append(f"- `{config_file}`")
        md.append("")

    if results.servers:
        md.append("## Discovered Servers")
        for server in results.servers:
            status = "DISABLED" if server.disabled else "ENABLED"
            md.append(f"### {server.name} ({server.type.value}) - {status}")
            if server.source_file:
                md.append(f"- **Source**: `{server.source_file}`")
            if server.endpoint:
                md.append(f"- **Endpoint**: `{server.endpoint}`")
            if server.command:
                md.append(f"- **Command**: `{server.command}`")
            if server.args:
                md.append(f"- **Args**: `{' '.join(_redact_args(server.args))}`")
            if server.headers:
                redacted_headers = _redact_headers(server.headers)
                headers_str = ', '.join([f"`{k}={v}`" for k, v in redacted_headers.items()])
                md.append(f"- **Headers**: {headers_str}")
            if server.env:
                redacted_env = _redact_env(server.env)
                env_str = ', '.join([f"`{k}={v}`" for k, v in redacted_env.items()])
                md.append(f"- **Environment Variables**: {env_str}")
            if server.additional_permissions:
                md.append(f"- **Permissions**: {', '.join(server.additional_permissions)}")
            md.append("")

    if results.tools:
        md.append("## Discovered Tools")
        for tool in results.tools:
            md.append(f"### {tool.name}")
            md.append(f"- **Server**: {tool.server_name}")
            md.append(f"- **Description**: {tool.description}")
            if tool.input_schema:
                md.append(f"- **Schema**:")
                md.append("```json")
                md.append(json.dumps(tool.input_schema, indent=2))
                md.append("```")
            md.append("")
    if results.resources:
        md.append("## Discovered Resources")
        for resource in results.resources:
            md.append(f"### {resource.name}")
            md.append(f"- **URI**: `{resource.uri}`")
            md.append(f"- **Server**: {resource.server_name}")
            md.append(f"- **Description**: {resource.description}")
            md.append(f"- **MIME Type**: {resource.mime_type}")
            md.append("")

    if results.prompts:
        md.append("## Discovered Prompts")
        for prompt in results.prompts:
            md.append(f"### {prompt.name}")
            md.append(f"- **Server**: {prompt.server_name}")
            md.append(f"- **Description**: {prompt.description}")
            if prompt.arguments:
                md.append(f"- **Arguments**:")
                md.append("```json")
                md.append(json.dumps(prompt.arguments, indent=2))
                md.append("```")
            md.append("")

    if results.resource_templates:
        md.append("## Discovered Resource Templates")
        for template in results.resource_templates:
            md.append(f"### {template.name}")
            md.append(f"- **URI Template**: `{template.uri_template}`")
            md.append(f"- **Server**: {template.server_name}")
            md.append(f"- **Description**: {template.description}")
            md.append(f"- **MIME Type**: {template.mime_type}")
            md.append("")

    if results.errors:
        md.append("## Errors")
        for error in results.errors:
            md.append(f"- {error}")
        md.append("")

    if results.status:
        md.append("## Status")
        for status_msg in results.status:
            md.append(f"- {status_msg}")
        md.append("")

    return "\n".join(md)


def determine_output_file(args, command_type: str):
    if args.output:
        return args.output
    else:
        return f"{command_type}_results.{args.report_type}"


def build_baseline_status(baseline_path: Optional[str], baseline_loaded: bool, baseline_warning: Optional[str]) -> dict:
    status = {
        "baseline_file": baseline_path if baseline_path else None,
        "baseline_loaded": baseline_loaded
    }
    if baseline_warning:
        status["warning"] = baseline_warning
    return status


def validate_baseline_path(file_path: str, mode: str) -> Optional[Path]:
    path = Path(file_path)
    
    if path.exists() and path.is_dir():
        raise ValueError(f"Baseline path is a directory: {file_path}. Must be a file path.")
    
    if mode == 'inspect':
        if not path.parent.exists():
            raise ValueError(
                f"Parent directory does not exist: {path.parent}\n"
                f"Please create the directory first or use an existing path."
            )
        return path
    
    if mode == 'scan':
        if not path.exists():
            return None
        if not path.is_file():
            raise ValueError(f"Baseline path exists but is not a file: {file_path}")
        return path
    
    return path


def save_report_to_file(content, filepath, report_type, suppress_print=False):
    output_path = Path(filepath)
    
    output_path.parent.mkdir(parents=True, exist_ok=True)
    
    if output_path.exists() and output_path.is_dir():
        raise ValueError(f"Output path is a directory: {filepath}. Must be a file path.")

    with open(output_path, 'w') as f:
        f.write(content)

    if not suppress_print:
        print(f"Report saved to {output_path}")


async def scan_command(args):
    print('\n🏃 Running MCP Armor in scan mode...\n')

    from .scanner import MCPConfigScanner
    from .security_utils import (
        ScanResult,
        AffectedEntity,
        calculate_risk_level,
        sort_issues_by_severity,
        group_issues_by_type_and_severity,
    )
    from .baseline import load_baseline, generate_baseline, save_baseline
    
    scanner = MCPConfigScanner()
    
    baseline = None
    baseline_path_str = None
    baseline_path = None
    baseline_warning = None
    
    if args.baseline:
        baseline_path_str = args.baseline
    else:
        default_baseline = Path("baseline.json")
        if default_baseline.exists():
            baseline_path_str = str(default_baseline)

    custom_paths = []
    if args.config:
        custom_paths.extend(args.config)
    
    if baseline_path_str:
        try:
            validated_path = validate_baseline_path(baseline_path_str, 'scan')
            if validated_path:
                baseline_path = validated_path
                baseline = load_baseline(validated_path)
                if baseline:
                    print(f"  🧭 Using baseline: {baseline_path_str} (✅ loaded)\n")
                else:
                    baseline_warning = f"Baseline checks skipped. Baseline file not found or invalid: {baseline_path_str}."
                    print(f"  ⚠️  Skipping baseline checks as no valid baseline file was provided: {baseline_path_str}.")
                    print(f"     💡 Run `mcp-armor inspect --baseline {baseline_path_str}` first to create one.")
            else:
                baseline_warning = f"Baseline checks skipped. Baseline file not found or invalid: {baseline_path_str}."
                print(f"  ⚠️  Skipping baseline checks as no valid baseline file was provided: {baseline_path_str}.")
                print(f"     💡 Run `mcp-armor inspect --baseline {baseline_path_str}` first to create one.\n")
        except ValueError as e:
            baseline_warning = f"Baseline checks skipped. Invalid baseline file: {baseline_path_str}."
            print(f"  ⚠️  Skipping baseline checks as no valid baseline file was provided: {baseline_path_str}.")
            print(f"     💡 Run `mcp-armor inspect --baseline {baseline_path_str}` first to create one.\n")
    else:
        baseline_warning = "Baseline checks skipped. No baseline file found."
        print(f"  ⚠️  Skipping baseline checks as no baseline file was found.")
        print(f"     💡 Run `mcp-armor inspect` first to create one.\n")
    missing_paths = []
    if not custom_paths:
        print("  📄 No config provided — searching known locations…")
    else:
        for p in custom_paths:
            exists = Path(p).exists()
            status_text = " (✅ found)" if exists else " (❌ not found)"
            print(f"  📄 Using config: {p}{status_text}")
            if not exists:
                missing_paths.append(p)
        if missing_paths:
            custom_paths = [p for p in custom_paths if Path(p).exists()]
            if not custom_paths:
                print(
                    "\n  No valid config files found. Please provide at least one valid config file path with --config.")
                sys.exit(1)

    try:
        with SuppressStderr():
            config_files = await scanner.discover_config_files(custom_paths)
            print("     🧩 Parsing configuration…")
            is_user_provided = bool(custom_paths)

            servers = []
            parse_errors = []
            for cfg in config_files:
                try:
                    parsed = await scanner.parse_single_config(cfg, is_user_provided)
                    servers.extend(parsed)
                except Exception as e:
                    parse_errors.append(f"Failed to parse config: {cfg} ({e})")

            if is_user_provided and len(servers) == 0:
                print("\n  ❌ Failed to parse provided config or no MCP servers found in it.")
                for p in config_files:
                    print(f"     ↳ {p}")
                if parse_errors:
                    for err in parse_errors:
                        print(f"     ❌ {err}")
                print("\n  Ensure the file is valid JSON/YAML and follows expected schema (servers/mcpServers).")
                sys.exit(1)

            server_names = [s.name for s in servers]
            if server_names:
                shown = ", ".join(server_names[:5])
                more = f" (+{len(server_names) - 5} more)" if len(server_names) > 5 else ""
                print(f"\n  🔲 Servers found ({len(server_names)}): {shown}{more}")

            failed = []
            seen_oauth = set()
            connected = set()

            def progress_cb(event, data):
                if event == 'failed':
                    name = data.get('server')
                    err = data.get('error', 'error')
                    failed.append((name, err))
                    clean_err = err.replace(f"Failed to connect to server: {name}", "").strip()
                    if not clean_err or clean_err == err:
                        clean_err = err
                    print(f"     ❌ Connection failed: {name} ({clean_err})")
                elif event == 'oauth_wait':
                    name = data.get('server')
                    if name and name not in seen_oauth:
                        seen_oauth.add(name)
                        print(f"        ⏳ Waiting for OAuth: {name}")
                elif event == 'oauth_approved':
                    name = data.get('server')
                    if name and getattr(args, 'verbose', False):
                        print(f"        🆗 Access approved: {name}")
                elif event == 'connected':
                    name = data.get('server')
                    if name:
                        connected.add(name)
                        tools_n = data.get('tools', 0)
                        res_n = data.get('resources', 0)
                        prompts_n = data.get('prompts', 0)
                        templates_n = data.get('resource_templates', 0)
                        print(f"        🔗 Connected: {name} (tools {tools_n}, resources {res_n}, prompts {prompts_n}, resource templates {templates_n})")

            from .run_checks import SecurityCheckRunner, SecurityReport
            security_scanner = SecurityCheckRunner(baseline=baseline)
            config_only_result = ScanResult(servers=servers, tools=[], resources=[], prompts=[], resource_templates=[], config_files=config_files)
            config_task = asyncio.create_task(security_scanner.run_configuration_checks(config_only_result))
            print("     🔌 Connecting to servers…")
            discovery_task = asyncio.create_task(scanner.discover_all(servers, progress=progress_cb))

            config_issues, (tools, resources, prompts, templates) = await asyncio.gather(config_task, discovery_task)

            errors = []
            status = []

            if missing_paths:
                for p in missing_paths:
                    errors.append(f"Config not found: {p}")
                if is_user_provided:
                    total_provided = len(args.config) if args.config else 0
                    missing_count = len(missing_paths)
                    status.append(
                        f"Some provided configs ({missing_count}/{total_provided}) not found. Results exclude those files.")

            if parse_errors:
                errors.extend(parse_errors)
                if is_user_provided:
                    total_cfgs = len(config_files)
                    failed_cfgs = len(parse_errors)
                    status.append(
                        f"Some provided configs ({failed_cfgs}/{total_cfgs}) failed to parse. Results exclude those files.")

            for server_name, error_msg in failed:
                clean_error = error_msg.replace(f"Failed to connect to server: {server_name}", "").strip()
                if not clean_error or clean_error == error_msg:
                    clean_error = error_msg
                errors.append(f"Connection failed to server '{server_name}': {clean_error}")

            enabled_servers = [s for s in servers if not s.disabled]
            total_servers = len(enabled_servers)
            connected_count = len(connected)
            failed_count = len(failed)

            if connected_count == 0 and failed_count > 0:
                status.append("No servers connected. Only configuration checks were run.")
            elif failed_count > 0:
                status.append(
                    f"Some servers ({failed_count}/{total_servers}) failed to connect. Security scan is incomplete for failed servers.")
            elif total_servers == 0:
                status.append("No MCP servers found in configuration. Only configuration checks were run.")
            else:
                if baseline:
                    status.append("All servers connected. Full security scan completed.")
                else:
                    status.append("All servers connected. Scan completed without baseline checks.")

            if len(connected) == 0:
                print("\n  ⚠️  No servers connected. Running configuration checks only.")
                print("     ‼️  Check your server configurations and network connectivity.")

            print("\n  ✅ Configuration scan complete\n")

            if len(connected) == 0:
                results = ScanResult(servers=servers, tools=[], resources=[], prompts=[], resource_templates=[], config_files=config_files, errors=errors,
                                     status=status)
                security_report = SecurityReport()
                security_report.all_issues = sort_issues_by_severity(config_issues)
                security_report.total_issues = len(config_issues)
                for issue in config_issues:
                    security_report.issues_by_severity[issue.severity] += 1
                    issue_type = issue.issue_type.split(':')[0]
                    if issue_type not in security_report.issues_by_type:
                        security_report.issues_by_type[issue_type] = 0
                    security_report.issues_by_type[issue_type] += 1
                security_report.grouped_issues = group_issues_by_type_and_severity(config_issues)
                security_report.risk_level = calculate_risk_level(config_issues)
                results.security_report = security_report
            else:
                print("  🛡️  Running security checks…")
                security_report = await security_scanner.perform_security_scan(
                    ScanResult(servers=servers, tools=tools, resources=resources, prompts=prompts, resource_templates=templates, config_files=config_files,
                               errors=errors, status=status)
                )
                results = ScanResult(servers=servers, tools=tools, resources=resources, prompts=prompts, resource_templates=templates, config_files=config_files,
                                     errors=errors, status=status)
                results.security_report = security_report

            output_file = determine_output_file(args, "scan")
            output_data = None
            markdown_content = None
            if args.report_type == "json":
                output_data = {
                    "servers": [
                        {
                            "name": server.name,
                            "type": server.type.value,
                            "endpoint": server.endpoint,
                            "command": server.command,
                            "args": _redact_args(server.args),
                            "headers": _redact_headers(server.headers),
                            "env": _redact_env(server.env),
                            "disabled": server.disabled,
                            "additional_permissions": server.additional_permissions,
                            "tools": [tool.name for tool in results.tools if tool.server_name == server.name],
                            "resources": [resource.name for resource in results.resources if
                                          resource.server_name == server.name],
                            "prompts": [prompt.name for prompt in results.prompts if prompt.server_name == server.name],
                            "resource_templates": [template.name for template in results.resource_templates if template.server_name == server.name],
                            "source_file": server.source_file
                        }
                        for server in results.servers
                    ],
                    "tools": [
                        {
                            "name": tool.name,
                            "description": tool.description,
                            "input_schema": tool.input_schema,
                            "server_name": tool.server_name,
                            "server_endpoint": tool.server_endpoint
                        }
                        for tool in results.tools
                    ],
                    "resources": [
                        {
                            "uri": resource.uri,
                            "name": resource.name,
                            "description": resource.description,
                            "mime_type": resource.mime_type,
                            "server_name": resource.server_name,
                            "server_endpoint": resource.server_endpoint
                        }
                        for resource in results.resources
                    ],
                    "prompts": [
                        {
                            "name": prompt.name,
                            "description": prompt.description,
                            "arguments": prompt.arguments,
                            "server_name": prompt.server_name,
                            "server_endpoint": prompt.server_endpoint
                        }
                        for prompt in results.prompts
                    ],
                    "resource_templates": [
                        {
                            "name": template.name,
                            "uri_template": template.uri_template,
                            "description": template.description,
                            "mime_type": template.mime_type,
                            "server_name": template.server_name,
                            "server_endpoint": template.server_endpoint
                        }
                        for template in results.resource_templates
                    ],
                    "config_files": results.config_files,
                    "errors": results.errors,
                    "status": results.status,
                    "baseline_status": build_baseline_status(baseline_path_str, baseline is not None, baseline_warning),
                    "security_report": {
                        "risk_level": results.security_report.risk_level.value if results.security_report else "NONE",
                        "total_findings": results.security_report.total_issues if results.security_report else 0,
                        "findings_by_severity": {k.value: v for k, v in (
                            results.security_report.issues_by_severity.items() if results.security_report else [])},
                        "findings_by_type": results.security_report.issues_by_type if results.security_report else {},
                        "grouped_findings": [
                            _grouped_issue_to_dict(grouped)
                            for grouped in (results.security_report.grouped_issues if results.security_report else [])
                        ]
                    }
                }
            else:
                baseline_status_dict = build_baseline_status(baseline_path_str, baseline is not None, baseline_warning)
                markdown_content = format_scan_results_as_markdown(results, baseline_status_dict)
            print()
            summary = scanner.get_scan_summary(results)
            print_separator('═', 60)
            print("📊 SUMMARY")
            print_separator('═', 60)
            print(f"🔲 Servers found:        {summary['total_servers']}")
            try:
                connected_count = len(connected)
                print(f"🔗 Connected servers:    {connected_count}/{summary['total_servers']}")
            except Exception:
                pass
            print(f"🔧 Tools found:          {summary['total_tools']}")
            print(f"📁 Resources found:      {summary['total_resources']}")
            print(f"💬 Prompts found:        {summary['total_prompts']}")
            print(f"🗂️  Resource Templates:   {summary['total_resource_templates']}")
            print(f"📄 Config files:         {summary['config_files']}")
            if baseline_path_str and baseline:
                print(f"🧭 Baseline Loaded:      Yes")
            else:
                print(f"🧭 Baseline Loaded:      No")
            if summary['errors'] > 0:
                print(f"❌ Errors:               {summary['errors']}")
            if results.status:
                print(f"💎 Status:")
                for msg in results.status:
                    print(f"   {msg}")
            print()
            print_separator('═', 60)
            risk = results.security_report.risk_level.value if results.security_report else "NONE"
            total_issues = results.security_report.total_issues if results.security_report else 0
            
            if total_issues == 0:
                icon = '✅'
            else:
                icon = get_level_emoji(risk, RISK_EMOJI)
            
            print(f"{icon} SECURITY REPORT")
            print_separator('═', 60)
            print(f"🔺 Total Findings:       {total_issues}")

            if results.security_report and total_issues > 0:
                print(f"\n📊 Findings by Severity:")
                severity_order = ['CRITICAL', 'HIGH', 'MEDIUM', 'LOW']
                for sev_name in severity_order:
                    for sev, count in results.security_report.issues_by_severity.items():
                        sev_str = sev.value if hasattr(sev, 'value') else str(sev)
                        if sev_str.upper() == sev_name and count > 0:
                            emoji = get_level_emoji(sev_str, SEVERITY_EMOJI)
                            print(f"   {emoji} {sev_str.upper()}: {count}")

                grouped_issues = results.security_report.grouped_issues
                if grouped_issues:
                    print(f"\n🪲  Findings by Type:\n")
                    
                    for grouped in grouped_issues:
                        sev_str = grouped.severity.value.upper()
                        emoji = get_level_emoji(sev_str, SEVERITY_EMOJI)
                        
                        print(f"   {emoji} {grouped.issue_type} ({sev_str})")
                        print(wrap_text_with_indent(grouped.description, indent="       ", max_lines=None))
                        
                        print(f"       🐞 Entities affected ({grouped.total_entities}):")
                        for entity in grouped.entities_affected:
                            entity_display = _format_entity_display(entity)
                            print(f"           ⮕  {entity_display}")
                        
                        print(f"       💡 Remediation:")
                        print(wrap_text_with_indent(grouped.remediation, indent="           ", max_lines=None))
                        print()

            try:
                if args.report_type == "json":
                    save_report_to_file(json.dumps(output_data, indent=2), output_file, "json", suppress_print=True)
                else:
                    save_report_to_file(markdown_content, output_file, "markdown", suppress_print=True)
                print(f"💾 Report saved to {output_file}")
            except Exception as e:
                print(f"\n⚠️  Warning: Failed to save report to {output_file}: {e}", file=sys.stderr)
    except Exception as e:
        print(f"Scan failed: {e}", file=sys.stderr)
        sys.exit(1)


async def inspect_command(args):
    print('\n🏃 Running MCP Armor in inspect mode...\n')

    from .scanner import MCPConfigScanner
    from .security_utils import ScanResult
    from .baseline import generate_baseline, save_baseline
    
    scanner = MCPConfigScanner()
    
    baseline_path_str = args.baseline if args.baseline else "baseline.json"
    baseline_path = None
    
    try:
        validated_path = validate_baseline_path(baseline_path_str, 'inspect')
        baseline_path = validated_path
    except ValueError as e:
        print(f"\n  ❌ Error: {e}")
        sys.exit(1)
    
    custom_paths = []
    if args.config:
        custom_paths.extend(args.config)
    
    missing_paths = []
    if not custom_paths:
        print("  📄 No config provided — searching known locations…")
    else:
        for p in custom_paths:
            exists = Path(p).exists()
            status_text = " (✅ found)" if exists else " (❌ not found)"
            print(f"  📄 Using config: {p}{status_text}")
            if not exists:
                missing_paths.append(p)
        if missing_paths:
            custom_paths = [p for p in custom_paths if Path(p).exists()]
            if not custom_paths:
                print(
                    "\n  No valid config files found. Please provide at least one valid config file path with --config.")
                sys.exit(1)

    try:
        with SuppressStderr():
            config_files = await scanner.discover_config_files(custom_paths)
            print("     🧩 Parsing configuration…")
            is_user_provided = bool(custom_paths)

            servers = []
            parse_errors = []
            for cfg in config_files:
                try:
                    parsed = await scanner.parse_single_config(cfg, is_user_provided)
                    servers.extend(parsed)
                except Exception as e:
                    parse_errors.append(f"Failed to parse config: {cfg} ({e})")

            if is_user_provided and len(servers) == 0:
                print("\n  ❌ Failed to parse provided config or no MCP servers found in it.")
                for p in config_files:
                    print(f"     ↳ {p}")
                if parse_errors:
                    for err in parse_errors:
                        print(f"     ❌ {err}")
                print("\n  Ensure the file is valid JSON/YAML and follows expected schema (servers/mcpServers).")
                sys.exit(1)

            server_names = [s.name for s in servers]
            if server_names:
                shown = ", ".join(server_names[:5])
                more = f" (+{len(server_names) - 5} more)" if len(server_names) > 5 else ""
                print(f"\n  🔲 Servers found ({len(server_names)}): {shown}{more}")

            connected = set()
            failed = []
            seen_connecting = set()
            seen_oauth = set()

            def progress_cb(event, data):
                if event == 'failed':
                    name = data.get('server')
                    err = data.get('error', 'error')
                    failed.append((name, err))
                    clean_err = err.replace(f"Failed to connect to server: {name}", "").strip()
                    if not clean_err or clean_err == err:
                        clean_err = err
                    print(f"     ❌ Connection failed: {name} ({clean_err})")
                elif event == 'oauth_wait':
                    name = data.get('server')
                    if name and name not in seen_oauth:
                        seen_oauth.add(name)
                        print(f"        ⏳ Waiting for OAuth: {name}")
                elif event == 'oauth_approved':
                    name = data.get('server')
                    if name and getattr(args, 'verbose', False):
                        print(f"        🆗 Access approved: {name}")
                elif event == 'connected':
                    name = data.get('server')
                    if name:
                        connected.add(name)
                        tools_n = data.get('tools', 0)
                        res_n = data.get('resources', 0)
                        prompts_n = data.get('prompts', 0)
                        templates_n = data.get('resource_templates', 0)
                        print(f"        🔗 Connected: {name} (tools {tools_n}, resources {res_n}, prompts {prompts_n}, resource templates {templates_n})")

            print("     🔌 Connecting to servers…")

            tools, resources, prompts, templates = await scanner.discover_all(servers, progress=progress_cb)

            errors = []
            status = []

            if missing_paths:
                for p in missing_paths:
                    errors.append(f"Config not found: {p}")
                if is_user_provided:
                    total_provided = len(args.config) if args.config else 0
                    missing_count = len(missing_paths)
                    status.append(
                        f"Some provided configs ({missing_count}/{total_provided}) not found. Results exclude those files.")

            if parse_errors:
                errors.extend(parse_errors)
                if is_user_provided:
                    total_cfgs = len(config_files)
                    failed_cfgs = len(parse_errors)
                    status.append(
                        f"Some provided configs ({failed_cfgs}/{total_cfgs}) failed to parse. Results exclude those files.")

            for server_name, error_msg in failed:
                clean_error = error_msg.replace(f"Failed to connect to server: {server_name}", "").strip()
                if not clean_error or clean_error == error_msg:
                    clean_error = error_msg
                errors.append(f"Connection failed to server '{server_name}': {clean_error}")

            enabled_servers = [s for s in servers if not s.disabled]
            total_servers = len(enabled_servers)
            connected_count = len(connected)
            failed_count = len(failed)

            if connected_count == 0 and (failed_count > 0 or total_servers > 0):
                status.append("No servers connected. Cannot proceed with inspection.")
                print("\n  ⚠️  No servers could be connected. Cannot proceed with inspection.")
                print("      ‼️  Check your server configurations and network connectivity.\n")
                results = ScanResult(servers=servers, tools=[], resources=[], prompts=[], resource_templates=[], config_files=config_files, errors=errors,
                                     status=status)
            elif total_servers == 0:
                status.append("No servers connected. Cannot proceed with inspection.")
                print("\n  ⚠️  No servers could be connected. Cannot proceed with inspection.")
                print("      ‼️  Check your server configurations and network connectivity.\n")
                results = ScanResult(servers=servers, tools=[], resources=[], prompts=[], resource_templates=[], config_files=config_files, errors=errors,
                                     status=status)
            elif failed_count > 0:
                status.append(
                    f"Some servers ({failed_count}/{total_servers}) failed to connect. Inspection results are incomplete for failed servers.")
                print("\n  🔍 Discovering tools/resources/prompts…\n")
                results = ScanResult(servers=servers, tools=tools, resources=resources, prompts=prompts, resource_templates=templates, config_files=config_files,
                                     errors=errors, status=status)
            else:
                status.append("All servers connected. Inspection complete.")
                print("\n  🔍 Discovering tools/resources/resource templates/prompts…\n")
                results = ScanResult(servers=servers, tools=tools, resources=resources, prompts=prompts, resource_templates=templates, config_files=config_files,
                                     errors=errors, status=status)

        baseline_data = generate_baseline(results)
        save_baseline(baseline_data, baseline_path)
        print(f"  🧭 Baseline saved to: {baseline_path_str}\n")

        output_file = determine_output_file(args, "inspect")
        output_data = None
        markdown_content = None

        if args.report_type == "json":
            output_data = {
                "servers": [
                    {
                        "name": server.name,
                        "type": server.type.value,
                        "endpoint": server.endpoint,
                        "command": server.command,
                        "args": server.args,
                        "headers": _redact_headers(server.headers),
                        "env": _redact_env(server.env),
                        "disabled": server.disabled,
                        "additional_permissions": server.additional_permissions,
                        "tools": [tool.name for tool in results.tools if tool.server_name == server.name],
                        "resources": [resource.name for resource in results.resources if
                                      resource.server_name == server.name],
                        "prompts": [prompt.name for prompt in results.prompts if prompt.server_name == server.name],
                        "resource_templates": [template.name for template in results.resource_templates if template.server_name == server.name],
                        "source_file": server.source_file
                    }
                    for server in results.servers
                ],
                "tools": [
                    {
                        "name": tool.name,
                        "description": tool.description,
                        "input_schema": tool.input_schema,
                        "server_name": tool.server_name,
                        "server_endpoint": tool.server_endpoint
                    }
                    for tool in results.tools
                ],
                "resources": [
                    {
                        "uri": resource.uri,
                        "name": resource.name,
                        "description": resource.description,
                        "mime_type": resource.mime_type,
                        "server_name": resource.server_name,
                        "server_endpoint": resource.server_endpoint
                    }
                    for resource in results.resources
                ],
                "prompts": [
                    {
                        "name": prompt.name,
                        "description": prompt.description,
                        "arguments": prompt.arguments,
                        "server_name": prompt.server_name,
                        "server_endpoint": prompt.server_endpoint
                    }
                    for prompt in results.prompts
                ],
                "resource_templates": [
                    {
                        "name": template.name,
                        "uri_template": template.uri_template,
                        "description": template.description,
                        "mime_type": template.mime_type,
                        "server_name": template.server_name,
                        "server_endpoint": template.server_endpoint
                    }
                    for template in results.resource_templates
                ],
                "config_files": results.config_files,
                "errors": results.errors,
                "status": results.status,
                "baseline": {
                    "baseline_file": baseline_path_str,
                    "baseline_generated": True,
                    "timestamp": baseline_data.timestamp
                }
            }

        else:
            baseline_info = {
                "baseline_file": baseline_path_str,
                "baseline_generated": True,
                "timestamp": baseline_data.timestamp
            }
            markdown_content = format_inspect_results_as_markdown(results, baseline_info)

        summary = scanner.get_scan_summary(results)
        print_separator('═', 60)
        print("📊 SUMMARY")
        print_separator('═', 60)
        print(f"🔲 Servers found:        {summary['total_servers']}")
        try:
            connected_count = len(connected)
            print(f"🔗 Connected servers:    {connected_count}/{summary['total_servers']}")
        except Exception:
            pass
        print(f"🔧 Tools found:          {summary['total_tools']}")
        print(f"📁 Resources found:      {summary['total_resources']}")
        print(f"💬 Prompts found:        {summary['total_prompts']}")
        print(f"🗂️  Resource Templates:   {summary['total_resource_templates']}")
        print(f"📄 Config files:         {summary['config_files']}")
        print(f"🧭 Baseline:             {baseline_path_str} (generated)")
        if summary['errors'] > 0:
            print(f"❌ Errors:               {summary['errors']}")
        if results.status:
            print(f"💎 Status:")
            for msg in results.status:
                print(f"   {msg}")
            print()
        print_separator('═', 60)
        print(f"📒 INSPECTION REPORT")
        print_separator('═', 60)
        print(f"📄 Configuration Files ({len(results.config_files)}):")
        for config_file in results.config_files:
            print(f"   • {config_file}")
        
        if results.servers:
            print(f"\n🔲  Servers ({len(results.servers)}):")
            for server in results.servers:
                status = "DISABLED" if server.disabled else "ENABLED"
                status_icon = "⛔" if server.disabled else "✅"
                print(f"   {status_icon} {server.name} ({server.type.value}) - {status}")
                if server.source_file:
                    print(f"      📍 Source: {server.source_file}")
                if server.endpoint:
                    print(f"      🔗 Endpoint: {server.endpoint}")
                if server.command:
                    print(f"      ⚙️ Command: {server.command}")
                if server.args:
                    print(f"      🔣 Args: {' '.join(_redact_args(server.args))}")
                if server.headers:
                    redacted_headers = _redact_headers(server.headers)
                    headers_str = ', '.join([f"{k}={v}" for k, v in redacted_headers.items()])
                    print(f"      📑 Headers: {headers_str}")
                if server.env:
                    redacted_env = _redact_env(server.env)
                    env_str = ', '.join([f"{k}={v}" for k, v in redacted_env.items()])
                    print(f"      🔐 Env: {env_str}")
                if server.additional_permissions:
                    print(f"      ⚠️  Permissions: {', '.join(server.additional_permissions)}")
                print()
        
        if results.tools:
            print(f"🔧 Tools ({len(results.tools)}):")
            for tool in results.tools:
                print(f"   • {tool.name}")
                print(f"     🔲 Server: {tool.server_name}")
                preview = truncate_lines(tool.description or "", 2).strip()
                if preview:
                    print(f"     📝 Description: {preview}\n")

        if results.resources:
            print(f"📁 Resources ({len(results.resources)}):")
            for resource in results.resources:
                print(f"   • {resource.name}")
                print(f"     🔲 Server: {resource.server_name}")
                print(f"     🔗 URI: {resource.uri}")
                if resource.description:
                    rprev = truncate_lines(resource.description, 2).strip()
                    print(f"     📝 Description: {rprev}")
                print(f"     📄 MIME Type: {resource.mime_type}\n")

        if results.prompts:
            print(f"💬 Prompts ({len(results.prompts)}):")
            for prompt in results.prompts:
                print(f"   • {prompt.name}")
                print(f"     🔲 Server: {prompt.server_name}")
                if prompt.description:
                    pprev = truncate_lines(prompt.description, 2).strip()
                    print(f"     📝 Description: {pprev}")
                if prompt.arguments:
                    arg_names = ', '.join(prompt.arguments.keys())
                    print(f"     🔣  Arguments: {arg_names} ({len(prompt.arguments)} defined)\n")

        if results.resource_templates:
            print(f"🗂️  Resource Templates ({len(results.resource_templates)}):")
            for template in results.resource_templates:
                print(f"   • {template.name}")
                print(f"     🔲 Server: {template.server_name}")
                print(f"     🔗 URI Template: {template.uri_template}")
                if template.description:
                    tprev = truncate_lines(template.description, 2).strip()
                    print(f"     📝 Description: {tprev}")
                print(f"     📄 MIME Type: {template.mime_type}\n")

        try:
            if args.report_type == "json":
                save_report_to_file(json.dumps(output_data, indent=2), output_file, "json", suppress_print=True)
            else:
                save_report_to_file(markdown_content, output_file, "markdown", suppress_print=True)
            print(f"💾 Report saved to {output_file}")
        except Exception as e:
            print(f"\n⚠️  Warning: Failed to save report to {output_file}: {e}", file=sys.stderr)

        if any("No servers connected" in status_msg for status_msg in results.status):
            sys.exit(1)
        
    except Exception as e:
        print(f"Inspection failed: {e}", file=sys.stderr)
        sys.exit(1)


def main():
    print_banner()
    parser = argparse.ArgumentParser(

        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog="""
Examples:
    ## Inspect Mode (Inventory and Generate Baseline)
        mcp-armor inspect                                 # Inspect all known configs and generate baseline.json (no security analysis)
        mcp-armor inspect --config ./config.json          # Inspect specific config and generate baseline.json (no security analysis)
        mcp-armor inspect --config ./config1.json \\
                            --config ./config2.json            # Inspect multiple configs and generate baseline.json (no security analysis)
        mcp-armor inspect --baseline ./my-baseline.json   # Generate custom baseline file (defaults to baseline.json if not specified)
        mcp-armor inspect --config ./config.json \\
                            --baseline ./my-baseline.json      # Inspect specific config and generate custom baseline file
        mcp-armor inspect --output results.json           # Save results to file
        mcp-armor inspect --verbose                       # Detailed output
    
    ## Scan Mode (Security Analysis)
        mcp-armor scan                                    # Scan all known configs using default baseline.json (if present) for security risks
        mcp-armor scan --config ./config.json             # Scan specific config using default baseline.json (if present)for security risks
        mcp-armor scan --config ./config1.json \\
                        --config ./config2.json                # Scan multiple configs using default baseline.json (if present) in one run
        mcp-armor scan --baseline ./my-baseline.json      # Scan all known configs with custom baseline for drift detection
        mcp-armor scan --config ./config.json \\
                        --baseline ./my-baseline.json          # Scan specific config with custom baseline for drift detection
        mcp-armor scan --output results.json              # Save results to file
        mcp-armor scan --verbose                          # Detailed output
        mcp-armor scan --show-logs                        # Display debug logs in terminal
    
Note: 
    First-time scan will download the ML model for prompt injection check (~290MB) from Hugging Face.
    Subsequent scans use the cached model for faster completion.
        """
    )
    
    subparsers = parser.add_subparsers(dest='command', help='Available commands')
    
    scan_parser = subparsers.add_parser(
        'scan', 
        help='Scan MCP configurations',
        description="""
        Scan MCP server configurations for detecting any security risks.
        
        Note: 
          On first run, the prompt injection detection model will be 
        automatically downloaded from Hugging Face (~290MB). 
          This is an one-time download and subsequent scans will use the cached model.
        """
    )
    scan_parser.add_argument('--config', action='append', help='Custom config file path')
    scan_parser.add_argument('--baseline', help='Baseline file path for comparison (default: baseline.json in current directory)')
    scan_parser.add_argument('--report-type', choices=['json', 'md'], default='json',
                             help='Output format: json or markdown (default: json)')
    scan_parser.add_argument('--output', '-o', help='Output file path (optional)')
    scan_parser.add_argument('--verbose', '-v', action='store_true', help='Verbose output')
    scan_parser.add_argument('--show-logs', action='store_true',
                             help='Display debug logs in terminal (logs always saved to file)')
    scan_parser.set_defaults(func=scan_command)
    
    inspect_parser = subparsers.add_parser(
        'inspect', 
        help='Inspect MCP configurations',
        description="""
        Inspect MCP server configurations and generate a baseline for future comparisons.
        """
    )
    inspect_parser.add_argument('--config', action='append', help='Custom config file path')
    inspect_parser.add_argument('--baseline', help='Baseline file path to save (default: baseline.json in current directory)')
    inspect_parser.add_argument('--report-type', choices=['json', 'md'], default='json',
                                help='Output format: json or markdown (default: json)')
    inspect_parser.add_argument('--output', '-o', help='Output file path (optional)')
    inspect_parser.add_argument('--verbose', '-v', action='store_true', help='Verbose output')
    inspect_parser.add_argument('--show-logs', action='store_true',
                                help='Display debug logs in terminal (logs always saved to file)')
    inspect_parser.set_defaults(func=inspect_command)
    
    args = parser.parse_args()
    
    if not args.command:
        parser.print_help()
        sys.exit(1)
    
    if hasattr(args, 'show_logs') and args.show_logs:
        setup_logging(show_logs=True)
    else:
        setup_logging(show_logs=False)

    asyncio.run(args.func(args))


if __name__ == '__main__':
    main()


import json
import logging
from typing import Dict, Any, Optional, List, Tuple
from pathlib import Path

import yaml

from .security_utils import safe_load_file, MAX_CONFIG_FILE_SIZE

logger = logging.getLogger(__name__)


def validate_mcp_config_structure(config: Dict[str, Any], file_path: str) -> List[str]:
    errors = []
    
    if not isinstance(config, dict):
        errors.append(f"Configuration file '{file_path}' must contain a JSON/YAML object")
        return errors
    
    has_servers = 'mcpServers' in config or 'servers' in config
    
    if not has_servers and not isinstance(config, dict):
        errors.append(f"Configuration file '{file_path}' must contain 'mcpServers' or 'servers' key, or be a server configuration object")
    
    if has_servers:
        servers_key = 'mcpServers' if 'mcpServers' in config else 'servers'
        servers = config.get(servers_key)
        
        if not isinstance(servers, dict):
            errors.append(f"'{servers_key}' in '{file_path}' must be an object/dictionary")
        elif len(servers) == 0:
            errors.append(f"'{servers_key}' in '{file_path}' is empty - no servers configured")
        else:
            for server_name, server_config in servers.items():
                if not isinstance(server_config, dict):
                    errors.append(f"Server '{server_name}' in '{file_path}' must be a configuration object")
                else:
                    if 'command' not in server_config and 'url' not in server_config and 'endpoint' not in server_config:
                        errors.append(
                            f"Server '{server_name}' in '{file_path}' must have at least one of: 'command', 'url', or 'endpoint'"
                        )
    
    return errors


def load_and_validate_config(file_path: str) -> Tuple[Optional[Dict[str, Any]], List[str]]:
    errors = []
    config = None
    
    if not Path(file_path).exists():
        errors.append(f"Configuration file '{file_path}' does not exist")
        return None, errors
    
    try:
        config = safe_load_file(Path(file_path), MAX_CONFIG_FILE_SIZE)
    except ValueError as e:
        errors.append(str(e))
        return None, errors
    except json.JSONDecodeError as e:
        errors.append(f"Invalid JSON in '{file_path}': {e}")
        return None, errors
    except yaml.YAMLError as e:
        errors.append(f"Invalid YAML in '{file_path}': {e}")
        return None, errors
    except Exception as e:
        errors.append(f"Error reading '{file_path}': {e}")
        return None, errors
    
    if config is None:
        errors.append(f"Configuration file '{file_path}' is empty")
        return None, errors
    
    validation_errors = validate_mcp_config_structure(config, file_path)
    errors.extend(validation_errors)
    
    return config, errors




import logging
from typing import List

from ..security_utils import (
    ScanResult,
    SecurityIssue,
    Severity,
    create_security_issue,
)
from ..baseline import (
    BaselineData,
    hash_tool_description,
    hash_tool_metadata,
    hash_tool_args,
    hash_resource_description,
    hash_resource_uri,
    hash_resource_mime_type,
    hash_prompt_description,
    hash_prompt_arguments,
    hash_prompt_metadata,
    hash_resource_template_description,
    hash_resource_template_uri_template,
    hash_resource_template_mime_type,
    hash_resource_template_metadata,
)

logger = logging.getLogger(__name__)

ISSUE_TYPE_RUG_PULL = "Rug Pull: MCP Component Modified"


async def scan_for_tool_rug_pull(scan_result: ScanResult, baseline: BaselineData) -> List[SecurityIssue]:
    issues = []
    
    if not baseline.tools:
        return issues
    
    baseline_tool_map = {
        (tool.server_name, tool.name, tool.config_file): tool
        for tool in baseline.tools
    }
    
    for tool in scan_result.tools:
        server = next((s for s in scan_result.servers if s.name == tool.server_name), None)
        config_file = server.source_file if server else ""
        
        tool_key = (tool.server_name, tool.name, config_file)
        baseline_tool = baseline_tool_map.get(tool_key)
        
        if not baseline_tool:
            continue
        
        changed_fields = []
        
        current_desc_hash = hash_tool_description(tool)
        if current_desc_hash != baseline_tool.description_hash:
            changed_fields.append("description")
        
        current_meta_hash = hash_tool_metadata(tool)
        if current_meta_hash != baseline_tool.metadata_hash:
            changed_fields.append("metadata")
        
        current_args_hash = hash_tool_args(tool)
        if current_args_hash != baseline_tool.args_hash:
            changed_fields.append("arguments")
        
        if changed_fields:
            issues.append(create_security_issue(
                issue_type=ISSUE_TYPE_RUG_PULL,
                severity=Severity.CRITICAL,
                entity_type="tool",
                affected_server=tool.server_name,
                affected_tool=tool.name,
                config_file=config_file,
                affected_entities={
                    "changed_fields": changed_fields
                }
            ))
    
    if issues:
        logger.info(f"Found {len(issues)} tool rug pull issues")
    
    return issues


async def scan_for_resource_rug_pull(scan_result: ScanResult, baseline: BaselineData) -> List[SecurityIssue]:
    issues = []
    
    if not baseline.resources:
        return issues
    
    baseline_resource_map = {
        (resource.server_name, resource.name, resource.config_file): resource
        for resource in baseline.resources
    }
    
    for resource in scan_result.resources:
        server = next((s for s in scan_result.servers if s.name == resource.server_name), None)
        config_file = server.source_file if server else ""
        
        resource_key = (resource.server_name, resource.name, config_file)
        baseline_resource = baseline_resource_map.get(resource_key)
        
        if not baseline_resource:
            continue
        
        changed_fields = []
        
        current_desc_hash = hash_resource_description(resource)
        if current_desc_hash != baseline_resource.description_hash:
            changed_fields.append("description")
        
        current_uri_hash = hash_resource_uri(resource)
        if current_uri_hash != baseline_resource.uri_hash:
            changed_fields.append("URI")
        
        current_mime_hash = hash_resource_mime_type(resource)
        if current_mime_hash != baseline_resource.mime_type_hash:
            changed_fields.append("MIME type")
        
        if changed_fields:
            issues.append(create_security_issue(
                issue_type=ISSUE_TYPE_RUG_PULL,
                severity=Severity.CRITICAL,
                entity_type="resource",
                affected_server=resource.server_name,
                affected_resource=resource.name,
                affected_resource_uri=resource.uri,
                config_file=config_file,
                affected_entities={
                    "changed_fields": changed_fields
                }
            ))
    
    if issues:
        logger.info(f"Found {len(issues)} resource rug pull issues")
    
    return issues


async def scan_for_prompt_rug_pull(scan_result: ScanResult, baseline: BaselineData) -> List[SecurityIssue]:
    issues = []
    
    if not baseline.prompts:
        return issues
    
    baseline_prompt_map = {
        (prompt.server_name, prompt.name, prompt.config_file): prompt
        for prompt in baseline.prompts
    }
    
    for prompt in scan_result.prompts:
        config_file = prompt.config_file or ""
        
        prompt_key = (prompt.server_name, prompt.name, config_file)
        baseline_prompt = baseline_prompt_map.get(prompt_key)
        
        if not baseline_prompt:
            continue
        
        changed_fields = []
        
        current_desc_hash = hash_prompt_description(prompt)
        if current_desc_hash != baseline_prompt.description_hash:
            changed_fields.append("description")
        
        current_args_hash = hash_prompt_arguments(prompt)
        if current_args_hash != baseline_prompt.arguments_hash:
            changed_fields.append("arguments")
        
        current_meta_hash = hash_prompt_metadata(prompt)
        if current_meta_hash != baseline_prompt.metadata_hash:
            changed_fields.append("metadata")
        
        if changed_fields:
            issues.append(create_security_issue(
                issue_type=ISSUE_TYPE_RUG_PULL,
                severity=Severity.CRITICAL,
                entity_type="prompt",
                affected_server=prompt.server_name,
                config_file=config_file,
                affected_entities={
                    "prompt": prompt.name,
                    "changed_fields": changed_fields
                }
            ))
    
    if issues:
        logger.info(f"Found {len(issues)} prompt rug pull issues")
    
    return issues


async def scan_for_resource_template_rug_pull(scan_result: ScanResult, baseline: BaselineData) -> List[SecurityIssue]:
    issues = []
    
    if not baseline.resource_templates:
        return issues
    
    baseline_template_map = {
        (template.server_name, template.name, template.config_file): template
        for template in baseline.resource_templates
    }
    
    for template in scan_result.resource_templates:
        config_file = template.config_file or ""
        
        template_key = (template.server_name, template.name, config_file)
        baseline_template = baseline_template_map.get(template_key)
        
        if not baseline_template:
            continue
        
        changed_fields = []
        
        current_desc_hash = hash_resource_template_description(template)
        if current_desc_hash != baseline_template.description_hash:
            changed_fields.append("description")
        
        current_uri_hash = hash_resource_template_uri_template(template)
        if current_uri_hash != baseline_template.uri_template_hash:
            changed_fields.append("URI template")
        
        current_mime_hash = hash_resource_template_mime_type(template)
        if current_mime_hash != baseline_template.mime_type_hash:
            changed_fields.append("MIME type")
        
        current_meta_hash = hash_resource_template_metadata(template)
        if current_meta_hash != baseline_template.metadata_hash:
            changed_fields.append("metadata")
        
        if changed_fields:
            issues.append(create_security_issue(
                issue_type=ISSUE_TYPE_RUG_PULL,
                severity=Severity.CRITICAL,
                entity_type="resource_template",
                affected_server=template.server_name,
                config_file=config_file,
                affected_entities={
                    "resource_template": template.name,
                    "uri_template": template.uri_template,
                    "changed_fields": changed_fields
                }
            ))
    
    if issues:
        logger.info(f"Found {len(issues)} resource template rug pull issues")
    
    return issues




import json
import logging
from pathlib import Path
from typing import List, Dict, Any, Optional, Tuple
from dataclasses import dataclass, field
from enum import Enum

logger = logging.getLogger(__name__)


MAX_CONFIG_FILE_SIZE = 10 * 1024 * 1024  # 10MB for config files
MAX_BASELINE_FILE_SIZE = 50 * 1024 * 1024  # 50MB for baseline files


class RiskLevel(Enum):
    NONE = "NONE"
    LOW = "LOW"
    MEDIUM = "MEDIUM"
    HIGH = "HIGH"
    CRITICAL = "CRITICAL"


class Severity(Enum):
    LOW = "low"
    MEDIUM = "medium"
    HIGH = "high"
    CRITICAL = "critical"


class TransportType(Enum):
    STDIO = "stdio"
    HTTP = "http"
    SSE = "sse"


@dataclass
class MCPServerInfo:
    name: str
    type: TransportType
    endpoint: Optional[str] = None
    command: Optional[str] = None
    args: List[str] = None
    headers: Dict[str, str] = None
    env: Dict[str, str] = None
    disabled: bool = False
    tools: List[str] = None
    resources: List[str] = None
    additional_permissions: List[str] = None
    source_file: Optional[str] = None
    
    def __post_init__(self):
        if self.args is None:
            self.args = []
        if self.headers is None:
            self.headers = {}
        if self.env is None:
            self.env = {}
        if self.tools is None:
            self.tools = []
        if self.resources is None:
            self.resources = []


@dataclass
class ResourceInfo:
    uri: str
    name: str
    description: str
    mime_type: str
    server_name: str
    server_endpoint: str
    config_file: str = ""
    title: Optional[str] = None
    tags: List[str] = field(default_factory=list)


@dataclass
class ToolInfo:
    name: str
    
    description: str = ""
    input_schema: Dict[str, Any] = field(default_factory=dict)
    server_name: str = ""
    server_endpoint: str = ""
    config_file: str = ""
    title: Optional[str] = None
    output_schema: Optional[Dict[str, Any]] = None
    tags: List[str] = field(default_factory=list)


@dataclass
class PromptInfo:
    name: str
    description: str
    arguments: Dict[str, Any] = field(default_factory=dict)
    server_name: str = ""
    server_endpoint: str = ""
    config_file: str = ""
    title: Optional[str] = None
    tags: List[str] = field(default_factory=list)


@dataclass
class ResourceTemplateInfo:
    uri_template: str
    name: str
    description: str
    mime_type: str
    server_name: str = ""
    server_endpoint: str = ""
    config_file: str = ""
    title: Optional[str] = None
    tags: List[str] = field(default_factory=list)


@dataclass
class ScanResult:
    servers: List[MCPServerInfo]
    tools: List[ToolInfo]
    resources: List[ResourceInfo]
    prompts: List[PromptInfo]
    resource_templates: List[ResourceTemplateInfo]
    config_files: List[str]
    errors: List[str] = field(default_factory=list)
    status: List[str] = field(default_factory=list)


@dataclass
class SecurityIssue:
    issue_type: str
    severity: Severity
    description: Optional[str] = None
    recommendation: Optional[str] = None

    entity_type: Optional[str] = None
    affected_server: Optional[str] = None
    config_file: Optional[str] = None
    affected_tool: Optional[str] = None
    affected_resource: Optional[str] = None
    affected_resource_uri: Optional[str] = None
    affected_entities: Optional[Dict[str, Any]] = None


@dataclass
class AffectedEntity:
    entity_type: str
    entity_name: str
    server_name: str
    config_file: Optional[str] = None
    
    affected_tool: Optional[str] = None
    affected_resource: Optional[str] = None
    affected_resource_uri: Optional[str] = None
    affected_prompt: Optional[str] = None
    affected_resource_template: Optional[str] = None
    affected_resource_template_uri: Optional[str] = None
    
    changed_fields: Optional[List[str]] = None
    additional_info: Optional[Dict[str, Any]] = None


@dataclass
class GroupedIssue:
    issue_type: str
    severity: Severity
    description: str
    remediation: str
    entities_affected: List[AffectedEntity] = field(default_factory=list)
    total_entities: int = 0


def create_security_issue(
        issue_type: str,
        severity: Severity,
        description: Optional[str] = None,
        recommendation: Optional[str] = None,
        entity_type: Optional[str] = None,
        affected_server: Optional[str] = None,
        config_file: Optional[str] = None,
        affected_tool: Optional[str] = None,
        affected_resource: Optional[str] = None,
        affected_resource_uri: Optional[str] = None,
        affected_entities: Optional[Dict[str, Any]] = None
) -> SecurityIssue:
    return SecurityIssue(
        issue_type=issue_type,
        severity=severity,
        description=description,
        recommendation=recommendation,
        entity_type=entity_type,
        affected_server=affected_server,
        config_file=config_file,
        affected_tool=affected_tool,
        affected_resource=affected_resource,
        affected_resource_uri=affected_resource_uri,
        affected_entities=affected_entities
    )


def extract_tool_content(tool: ToolInfo) -> str:
    tool_data = {
        "name": tool.name,
        "title": getattr(tool, 'title', None),
        "description": tool.description,
        "input_schema": tool.input_schema,
        "output_schema": getattr(tool, 'output_schema', None),
        "tags": getattr(tool, 'tags', None) or [],
        "server_name": tool.server_name,
        "server_endpoint": tool.server_endpoint
    }
    return json.dumps(tool_data, indent=2)


def extract_resource_content(resource: ResourceInfo) -> str:
    resource_data = {
        "uri": resource.uri,
        "name": resource.name,
        "title": getattr(resource, 'title', None),
        "description": resource.description,
        "mime_type": resource.mime_type,
        "tags": getattr(resource, 'tags', None) or [],
        "server_name": resource.server_name,
        "server_endpoint": resource.server_endpoint
    }
    return json.dumps(resource_data, indent=2)


def extract_prompt_content(prompt: PromptInfo) -> str:
    prompt_data = {
        "name": prompt.name,
        "title": getattr(prompt, 'title', None),
        "description": prompt.description,
        "arguments": prompt.arguments,
        "tags": getattr(prompt, 'tags', None) or [],
        "server_name": prompt.server_name,
        "server_endpoint": prompt.server_endpoint
    }
    return json.dumps(prompt_data, indent=2)


def extract_resource_template_content(template: ResourceTemplateInfo) -> str:
    template_data = {
        "uri_template": template.uri_template,
        "name": template.name,
        "title": getattr(template, 'title', None),
        "description": template.description,
        "mime_type": template.mime_type,
        "tags": getattr(template, 'tags', None) or [],
        "server_name": template.server_name,
        "server_endpoint": template.server_endpoint
    }
    return json.dumps(template_data, indent=2)


def extract_text_content(content: str) -> str:
    try:
        data = json.loads(content)
        text_parts = []

        if isinstance(data, dict):
            for key in ['description', 'name', 'uri', 'uri_template', 'title', 'content']:
                if key in data and data[key]:
                    text_parts.append(str(data[key]))

            if 'input_schema' in data and isinstance(data['input_schema'], dict):
                schema_str = json.dumps(data['input_schema'], indent=2)
                text_parts.append(schema_str)

            if 'output_schema' in data and isinstance(data['output_schema'], dict):
                schema_str = json.dumps(data['output_schema'], indent=2)
                text_parts.append(schema_str)

            if 'arguments' in data and isinstance(data['arguments'], dict):
                arguments_str = json.dumps(data['arguments'], indent=2)
                text_parts.append(arguments_str)

            if 'tags' in data and isinstance(data['tags'], list):
                tags_str = ' '.join(str(tag) for tag in data['tags'] if tag)
                if tags_str:
                    text_parts.append(tags_str)

        return ' '.join(text_parts)
    except (json.JSONDecodeError, TypeError):
        return content


def _severity_order(severity: Severity) -> int:
    order = {
        Severity.CRITICAL: 0,
        Severity.HIGH: 1,
        Severity.MEDIUM: 2,
        Severity.LOW: 3,
    }
    return order.get(severity, 99)


def sort_issues_by_severity(issues: List[SecurityIssue]) -> List[SecurityIssue]:
    return sorted(
        issues,
        key=lambda issue: (
            _severity_order(issue.severity),
            issue.issue_type or "",
            issue.affected_server or ""
        )
    )


def get_generic_description(issue_type: str) -> str:
    descriptions = {
        "Prompt Injection": "MCP components with hidden instructions may alter agent behavior and trigger unintended or malicious actions.",
        "Command Injection": "Hidden command execution patterns found in MCP component metadata, indicating potential command injection vectors.",
        "Hardcoded Secret": "Potential hardcoded secrets detected in MCP configuration file(s).",
        "Tool Name Ambiguity": "Tool names are very similar, which may cause agent misselection and unintended behavior.",
        "Cross-Server Tool Shadowing": "Tools reference other tools across server boundaries, potentially bypassing trust controls and causing agent misselection.",
        "Excessive Tool Permissions": "Excessive host permissions detected for MCP tools on the affected server.",
        "Rug Pull: MCP Component Modified": "The MCP component has changed since baseline approval. This change could enable data exfiltration or cause the agent to perform malicious actions.",
    }
    return descriptions.get(issue_type, f"{issue_type} detected.")


def get_generic_recommendation(issue_type: str) -> str:
    recommendations = {
        "Prompt Injection": (
            "Inspect affected MCP components and remove hidden instructions or block the component "
            "to prevent unauthorized actions and manipulation of the agent’s context."
        ),
        "Command Injection": (
            "Integrate only trusted MCP servers; block components that embed hidden or unsafe commands in metadata. "
            "Regularly scan for poisoning and rug-pull activity to prevent unauthorized execution or data exfiltration."
        ),
        "Hardcoded Secret": (
            "Remove hardcoded secrets and use environment variables or a secrets manager."
        ),
        "Tool Name Ambiguity": (
            "Isolate agents for conflicting MCP servers, or add guardrails to agent decision logic for safe tool selection."
        ),
        "Cross-Server Tool Shadowing": (
            "Connect only to trusted MCP servers. Enforce strict namespace isolation by assigning each tool a unique prefix "
            "based on its originating server. Implement guardrails to detect and block suspicious cross-server references or manipulation."
        ),
        "Excessive Tool Permissions": (
            "Enforce least privilege by restricting MCP server privileges on the host. Remove wildcard or admin-level permissions and disable terminal or file-system access flags. "
        ),
        "Rug Pull: MCP Component Modified": (
            "Review the modification. If suspicious, revert to the last safe MCP server version or block it. If expected, rerun inspect to update the baseline."
        ),
    }
    return recommendations.get(issue_type, f"Review and address {issue_type} issues.")


def _extract_entity_name(issue: SecurityIssue) -> str:
    if issue.affected_entities:
        if "prompt" in issue.affected_entities:
            return issue.affected_entities["prompt"]
        elif "resource_template" in issue.affected_entities:
            return issue.affected_entities["resource_template"]
    
    if issue.affected_tool:
        return issue.affected_tool
    elif issue.affected_resource:
        return issue.affected_resource
    elif issue.affected_resource_uri:
        return issue.affected_resource_uri
    elif issue.affected_server:
        return issue.affected_server
    
    logger.warning(
        f"SecurityIssue '{issue.issue_type}' has no identifiable entity. "
        f"Missing: affected_tool={issue.affected_tool}, affected_resource={issue.affected_resource}, "
        f"affected_resource_uri={issue.affected_resource_uri}, affected_server={issue.affected_server}, "
        f"affected_entities={issue.affected_entities}"
    )
    return "unknown"


def _extract_entity_from_issue(issue: SecurityIssue) -> AffectedEntity:
    entity_name = _extract_entity_name(issue)
    
    
    entity = AffectedEntity(
        entity_type=issue.entity_type or "unknown",
        entity_name=entity_name,
        server_name=issue.affected_server or "",
        config_file=issue.config_file,
        affected_tool=issue.affected_tool,
        affected_resource=issue.affected_resource,
        affected_resource_uri=issue.affected_resource_uri,
        affected_prompt=issue.affected_entities.get("prompt") if issue.affected_entities else None,
        affected_resource_template=issue.affected_entities.get("resource_template") if issue.affected_entities else None,
        affected_resource_template_uri=issue.affected_entities.get("uri_template") if issue.affected_entities else None,
        changed_fields=issue.affected_entities.get("changed_fields") if issue.affected_entities else None,
        additional_info=issue.affected_entities.copy() if issue.affected_entities else None
    )
    
    if entity.additional_info:
        entity.additional_info.pop("prompt", None)
        entity.additional_info.pop("resource_template", None)
        entity.additional_info.pop("uri_template", None)
        entity.additional_info.pop("changed_fields", None)
        if not entity.additional_info:
            entity.additional_info = None
    
    return entity


def group_issues_by_type_and_severity(issues: List[SecurityIssue]) -> List[GroupedIssue]:
    grouped_dict: Dict[Tuple[str, Severity], GroupedIssue] = {}
    
    for issue in issues:
        key = (issue.issue_type, issue.severity)
        
        if key not in grouped_dict:
            grouped_dict[key] = GroupedIssue(
                issue_type=issue.issue_type,
                severity=issue.severity,
                description=get_generic_description(issue.issue_type),
                remediation=get_generic_recommendation(issue.issue_type),
                entities_affected=[],
                total_entities=0
            )
        
        entity = _extract_entity_from_issue(issue)
        grouped_dict[key].entities_affected.append(entity)
        grouped_dict[key].total_entities += 1
    
    return sorted(
        grouped_dict.values(),
        key=lambda g: (_severity_order(g.severity), g.issue_type)
    )


def calculate_risk_level(issues: List[SecurityIssue]) -> RiskLevel:
    if not issues:
        return RiskLevel.NONE

    highest_severity = max(issues, key=lambda issue: _severity_order(issue.severity)).severity

    severity_to_risk = {
        Severity.CRITICAL: RiskLevel.CRITICAL,
        Severity.HIGH: RiskLevel.HIGH,
        Severity.MEDIUM: RiskLevel.MEDIUM,
        Severity.LOW: RiskLevel.LOW
    }

    return severity_to_risk[highest_severity]


def _check_file_size(file_path: Path, max_size: int) -> None:
    file_size = file_path.stat().st_size
    if file_size > max_size:
        raise ValueError(
            f"File '{file_path}' exceeds size limit ({max_size / (1024*1024):.1f}MB). "
            f"File size: {file_size / (1024*1024):.1f}MB"
        )


def safe_load_file(file_path: Path, max_size: int = MAX_CONFIG_FILE_SIZE) -> Dict[str, Any]:
    _check_file_size(file_path, max_size)
    
    with open(file_path, 'r', encoding='utf-8') as f:
        if str(file_path).endswith(('.yaml', '.yml')):
            import yaml
            return yaml.safe_load(f)
        else:
            return json.load(f)


import asyncio
import json
import re
from typing import Dict, List, Set, Tuple

from ..security_utils import (
    ScanResult,
    ToolInfo,
    SecurityIssue,
    Severity,
    create_security_issue,
    extract_tool_content,
)


def _normalize(name: str) -> str:
    return (name or "").strip().lower()


def _build_patterns_with_names(name_set: Set[str], exclude: Set[str]) -> List[Tuple[re.Pattern, str]]:
    patterns = []
    for name in name_set:
        norm = _normalize(name)
        if norm not in exclude and len(name.strip()) >= 3:
            escaped = re.escape(name)
            patterns.append((re.compile(rf"\b{escaped}\b", re.IGNORECASE), name))
    return patterns


def _scan_single_tool(
        tool: ToolInfo,
        tool_patterns: List[Tuple[re.Pattern, str]],
        server_patterns: List[Tuple[re.Pattern, str]],
        tool_to_servers: Dict[str, Set[str]],
) -> Tuple[Set[str], Set[str]]:
    content = extract_tool_content(tool)
    ref_tools = set()
    ref_servers = set()

    for pattern, orig_tool_name in tool_patterns:
        referenced_servers = tool_to_servers.get(orig_tool_name, set())

        if not referenced_servers:
            continue

        if orig_tool_name == tool.name:
            tool_data = {
                "description": tool.description,
                "input_schema": tool.input_schema,
                "server_name": tool.server_name,
                "server_endpoint": tool.server_endpoint
            }
            content_without_name = json.dumps(tool_data, indent=2)
            if not pattern.search(content_without_name):
                continue
        elif not pattern.search(content):
            continue

        if orig_tool_name != tool.name and tool.server_name and tool.server_name not in referenced_servers:
            ref_tools.add(orig_tool_name)
        elif orig_tool_name == tool.name and len(referenced_servers) > 1:
            mentioned_cross_servers = set()
            for srv in referenced_servers:
                if srv and srv != tool.server_name:
                    if re.search(rf"\b{re.escape(srv)}\b", content, re.IGNORECASE):
                        mentioned_cross_servers.add(srv)
            if mentioned_cross_servers:
                ref_tools.add(orig_tool_name)

    for pattern, orig_server_name in server_patterns:
        if pattern.search(content) and orig_server_name != tool.server_name:
            ref_servers.add(orig_server_name)

    return ref_tools, ref_servers


async def scan_for_cross_server_tool_shadowing(scan_result: ScanResult) -> List[SecurityIssue]:
    if not scan_result.tools:
        return []

    tool_names: Dict[str, str] = {}
    tool_to_servers: Dict[str, Set[str]] = {}
    server_names: Set[str] = set()

    for tool in scan_result.tools:
        tool_names[_normalize(tool.name)] = tool.name
        if tool.name and tool.server_name:
            if tool.name not in tool_to_servers:
                tool_to_servers[tool.name] = set()
            tool_to_servers[tool.name].add(tool.server_name)
        if tool.server_name:
            server_names.add(tool.server_name)

    for server in scan_result.servers:
        if server.name:
            server_names.add(server.name)

    issues: List[SecurityIssue] = []

    for tool in scan_result.tools:
        exclude_servers = {_normalize(tool.server_name)} if tool.server_name else set()

        tool_patterns = _build_patterns_with_names(
            set(tool_names.values()), set()
        )
        server_patterns = _build_patterns_with_names(server_names, exclude_servers)

        ref_tools, ref_servers = await asyncio.to_thread(
            _scan_single_tool, tool, tool_patterns, server_patterns, tool_to_servers
        )

        if ref_tools or ref_servers:
            severity = Severity.HIGH if ref_tools else Severity.MEDIUM
            tool_list = sorted(ref_tools)
            server_list = sorted(ref_servers)

            parts = []
            if tool_list:
                parts.append(f"tools {tool_list}")
            if server_list:
                parts.append(f"servers {server_list}")

            contexts_str = ", ".join(parts)
            count_parts = []
            if ref_tools:
                count_parts.append(f"{len(ref_tools)} cross tool name reference{'s' if len(ref_tools) != 1 else ''}")
            if ref_servers:
                count_parts.append(
                    f"{len(ref_servers)} cross server name reference{'s' if len(ref_servers) != 1 else ''}")

            server_obj = next(
                (s for s in scan_result.servers if s.name == tool.server_name), None
            )
            config_file = server_obj.source_file if server_obj else None

            issues.append(
                create_security_issue(
                    issue_type="Cross-Server Tool Shadowing",
                    severity=severity,
                    entity_type="tool",
                    affected_tool=tool.name,
                    affected_server=tool.server_name,
                    config_file=config_file,
                    affected_entities={
                        "referenced_tools": sorted(ref_tools),
                        "referenced_servers": sorted(ref_servers),
                    }
                )
            )

    return issues


import os
import json
import yaml
import asyncio
import logging
from typing import Dict, List, Optional, Any, Tuple, Callable
from pathlib import Path
from .config_validator import load_and_validate_config
from .connector import MCPConnector
from .security_utils import (
    TransportType,
    MCPServerInfo,
    ToolInfo,
    ResourceInfo,
    PromptInfo,
    ResourceTemplateInfo,
    ScanResult
)

logger = logging.getLogger(__name__)


def _safe_progress_callback(progress, event: str, data: Dict[str, Any]):
    if progress:
        try:
            progress(event, data)
        except Exception:
            pass


class MCPConfigScanner:
    
    def __init__(self):
        self.known_locations = {
            # Cursor
            "~/.cursor/mcp.json": "cursor",
            # Windsurf
            "~/.codeium/windsurf/mcp_config.json": "windsurf",
            # VS Code
            "~/.vscode/mcp.json": "vscode",
            "~/AppData/Roaming/Code/User/settings.json": "vscode",
            "~/AppData/Roaming/Code/User/mcp.json": "vscode",
            "~/Library/Application Support/Code/User/settings.json": "vscode",
            "~/Library/Application Support/Code/User/mcp.json": "vscode",
            "~/.config/Code/User/settings.json": "vscode",
            "~/.config/Code/User/mcp.json": "vscode",
            # Claude Desktop (macOS)
            "~/Library/Application Support/Claude/claude_desktop_config.json": "claude_desktop",
            # Claude Code
            "~/.claude/settings.json": "claude_desktop",
            "~/AppData/Roaming/Claude/claude_desktop_config.json": "claude_desktop",
            # Gemini CLI
            "~/.gemini/settings.json": "gemini_cli",
            ".gemini/settings.json": "gemini_cli",
            # Neovim
            "~/.config/nvim/mcp.json": "neovim",
            # Helix
            "~/.config/helix/mcp.json": "helix",
            # Zed
            "~/.config/zed/mcp.json": "zed",
            # Fallback locations
            "./mcp_servers.json": "custom",
            "./mcp.json": "custom",
            "~/.config/mcp/servers.json": "custom"
        }
    
    async def discover_config_files(self, custom_paths: List[str] = None) -> List[str]:
        config_files = []
        
        if custom_paths:
            config_files.extend(custom_paths)
            logger.info(f"Using custom config files: {custom_paths}")
        else:
            for location in self.known_locations:
                expanded_path = os.path.expanduser(location)
                if os.path.exists(expanded_path):
                    config_files.append(expanded_path)
                    logger.info(f"Found config file: {expanded_path}")
        
        if not custom_paths:
            current_dir = Path(".")
            for pattern in ["mcp*.json", "*.mcp.json", "mcp*.yaml", "*.mcp.yaml"]:
                for file_path in current_dir.glob(pattern):
                    if file_path.is_file():
                        config_files.append(str(file_path))
                        logger.info(f"Found config file: {file_path}")
        
        self.discovered_configs = config_files
        return config_files
    
    async def parse_config_files(self, config_files: List[str], is_user_provided: bool = False) -> List[MCPServerInfo]:
        servers = []
        
        for config_file in config_files:
            try:
                all_servers = await self.parse_single_config(config_file, is_user_provided)
                servers.extend(all_servers)
                logger.info(f"Parsed {len(all_servers)} servers from {config_file}")
            except Exception as e:
                logger.error(f"Failed to parse {config_file}: {e}")
        
        return servers
    
    async def parse_single_config(self, config_file: str, is_user_provided: bool = False) -> List[MCPServerInfo]:
        
        
        config, validation_errors = load_and_validate_config(config_file)
        
        if config is None:
            if validation_errors:
                error_msg = "; ".join(validation_errors)
                raise ValueError(f"Configuration validation failed for '{config_file}': {error_msg}")
            raise ValueError(f"Failed to load configuration file: {config_file}")
        
        if validation_errors and is_user_provided:
            for error in validation_errors:
                logger.warning(f"Configuration validation warning: {error}")
        
        servers = []
        
        if 'mcpServers' in config:
            mcp_servers = config['mcpServers']
        elif 'servers' in config:
            mcp_servers = config['servers']
        else:
            mcp_servers = config
        
        for server_name, server_config in mcp_servers.items():
            try:
                server_info = self.create_server_info(server_name, server_config, config_file, is_user_provided)
                servers.append(server_info)
            except Exception as e:
                logger.error(f"Failed to parse server {server_name}: {e}", exc_info=False)
        
        return servers
    
    def create_server_info(self, name: str, config: Dict[str, Any], source_file: str,
                           is_user_provided: bool = False) -> MCPServerInfo:
        server_type = config.get('transport') or config.get('type')

        if not server_type:
            if config.get('url') or config.get('endpoint'):
                server_type = 'http'
            elif '--transport' in config.get('args', []) and 'http-only' in config.get('args', []):
                server_type = 'http'
            else:
                server_type = 'stdio'
                logger.warning(f"Server '{name}' missing 'type' field, so defaulting to 'stdio' "
                             f"Consider explicitly setting the 'type' field in the configuration.")

        try:
            transport_type = TransportType(server_type.lower())
        except ValueError:
            raise ValueError(
                f"Server '{name}' has invalid transport type '{server_type}'. Valid types: {[t.value for t in TransportType]}")

        endpoint = config.get('url') or config.get('endpoint')
        command = config.get('command')
        args = config.get('args', [])
        headers = config.get('headers', {})
        env = config.get('env', {})
        disabled = config.get('disabled', False)

        additional_permissions = None
        if not is_user_provided:
            client_type = self._detect_client_type(source_file)
            if client_type != 'custom':
                additional_permissions = self._extract_client_permissions_from_config(config, source_file)

        server_info = MCPServerInfo(
            name=name,
            type=transport_type,
            endpoint=endpoint,
            command=command,
            args=args,
            headers=headers,
            env=env,
            disabled=disabled,
            tools=[],
            resources=[],
            source_file=source_file
        )
    
        if additional_permissions is not None:
            server_info.additional_permissions = additional_permissions

        return server_info

    def _extract_client_permissions_from_config(self, config: Dict[str, Any], source_file: str) -> List[str]:
        client_type = self._detect_client_type(source_file)

        client_permissions = []

        if client_type == 'windsurf':
            windsurf_permissions = ['codebase_search', 'find', 'grep_search', 'list_directory',
                                    'read_file', 'edit_file', 'write_to_file', 'run_terminal_command']
            for permission in windsurf_permissions:
                if permission in config:
                    client_permissions.append(permission)

        elif client_type == 'cursor':
            cursor_permissions = ['Read File', 'List Directory', 'Codebase', 'Grep',
                                  'Search Files', 'Web', 'Fetch Rules', 'Edit & Reapply',
                                  'Delete File', 'Terminal']
            for permission in cursor_permissions:
                if permission in config:
                    client_permissions.append(permission)

        elif client_type == 'vscode':
            vscode_permissions = ['extensions', 'fetch', 'findTestFiles', 'githubRepo',
                                  'new', 'openSimpleBrowser', 'problems', 'runCommands',
                                  'runNotebooks', 'runTasks', 'search', 'searchResults',
                                  'terminalLastCommand', 'terminalSelection', 'testFailure',
                                  'usages', 'vscodeAPI', 'changes', 'codebase', 'editFiles']
            for permission in vscode_permissions:
                if permission in config:
                    client_permissions.append(permission)

        elif client_type == 'custom':
            standard_keys = {'type', 'command', 'args', 'headers', 'env', 'disabled', 'name'}
            for key, value in config.items():
                if key not in standard_keys and isinstance(value, bool) and value:
                    client_permissions.append(key)

        return client_permissions

    def _detect_client_type(self, source_file: str) -> str:
        import os

        expanded_path = os.path.expanduser(source_file)

        for known_path, client_type in self.known_locations.items():
            expanded_known_path = os.path.expanduser(known_path)
            if expanded_path == expanded_known_path:
                return client_type

        return 'custom'

    async def discover_all(self, servers: List[MCPServerInfo], progress: Optional[Callable[[str, Dict[str, Any]], None]] = None) -> Tuple[
        List[ToolInfo], List[ResourceInfo], List[PromptInfo], List[ResourceTemplateInfo]]:
        connector = MCPConnector()
        all_tools: List[ToolInfo] = []
        all_resources: List[ResourceInfo] = []
        all_prompts: List[PromptInfo] = []
        all_resource_templates: List[ResourceTemplateInfo] = []
        
        async def discover_for_server(server: MCPServerInfo) -> Tuple[
            List[ToolInfo], List[ResourceInfo], List[PromptInfo], List[ResourceTemplateInfo], str, Optional[str]]:
            if server.disabled:
                logger.info(f"Skipping disabled server: {server.name}")
                return [], [], [], [], server.name, None
            try:
                _safe_progress_callback(progress, 'connecting', {"server": server.name})
                server_tools, server_resources, server_prompts, server_resource_templates = await connector.discover_all(server, progress=progress)
                logger.info(
                    f"Discovered {len(server_tools)} tools, {len(server_resources)} resources, {len(server_prompts)} prompts, {len(server_resource_templates)} resource templates from {server.name}")
                _safe_progress_callback(progress, 'connected', {
                    "server": server.name,
                    "tools": len(server_tools),
                    "resources": len(server_resources),
                    "prompts": len(server_prompts),
                    "resource_templates": len(server_resource_templates)
                })
                return server_tools, server_resources, server_prompts, server_resource_templates, server.name, None
            except Exception as e:
                logger.error(f"Failed to discover from {server.name}: {e}", exc_info=False)
                _safe_progress_callback(progress, 'failed', {"server": server.name, "error": str(e)})
                return [], [], [], [], server.name, str(e)

        tasks = [asyncio.create_task(discover_for_server(s)) for s in servers]
        results = await asyncio.gather(*tasks, return_exceptions=False)

        for tools, resources, prompts, templates, _srv, _err in results:
            all_tools.extend(tools)
            all_resources.extend(resources)
            all_prompts.extend(prompts)
            all_resource_templates.extend(templates)

        return all_tools, all_resources, all_prompts, all_resource_templates
    
    def get_scan_summary(self, scan_result: ScanResult) -> Dict[str, Any]:
        if not scan_result:
            return {"error": "No scan results available"}
        
        return {
            "total_servers": len(scan_result.servers),
            "total_tools": len(scan_result.tools),
            "total_resources": len(scan_result.resources),
            "total_prompts": len(scan_result.prompts),
            "total_resource_templates": len(scan_result.resource_templates),
            "config_files": len(scan_result.config_files),
            "errors": len(scan_result.errors),
            "status": len(scan_result.status),
            "servers_by_type": {
                transport.value: len([s for s in scan_result.servers if s.type == transport])
                for transport in TransportType
            }
        }